from asdf.extension import Converter


class NDCollectionConverter(Converter):
    tags = ["tag:sunpy.org:ndcube/ndcollection-*"]
    types = ["ndcube.ndcollection.NDCollection"]

    def from_yaml_tree(self, node, tag, ctx):
        from ndcube.ndcollection import NDCollection

        aligned_axes = list(node.get("aligned_axes", {}).values())
        aligned_axes = tuple(tuple(lst) for lst in aligned_axes)
        return NDCollection(node["items"], meta=node.get("meta"), aligned_axes=aligned_axes)

    def to_yaml_tree(self, ndcollection, tag, ctx):
        node = {}
        node["items"] = dict(ndcollection)
        if ndcollection.meta is not None:
            node["meta"] = ndcollection.meta
        if ndcollection._aligned_axes is not None:
            node["aligned_axes"] = ndcollection._aligned_axes

        return node



from asdf.extension import Converter


class CompoundConverter(Converter):
    tags = ["tag:sunpy.org:ndcube/compoundwcs-*"]
    types = ["ndcube.wcs.wrappers.compound_wcs.CompoundLowLevelWCS"]

    def from_yaml_tree(self, node, tag, ctx):
        from ndcube.wcs.wrappers import CompoundLowLevelWCS

        return CompoundLowLevelWCS(*node["wcs"], mapping=node.get("mapping"), pixel_atol=node.get("atol"))

    def to_yaml_tree(self, compoundwcs, tag, ctx):
        node = {}
        node["wcs"] = compoundwcs._wcs
        node["mapping"] = compoundwcs.mapping.mapping
        node["atol"] = compoundwcs.atol
        return node



from asdf.extension import Converter


class ExtraCoordsConverter(Converter):
    tags = ["tag:sunpy.org:ndcube/extra_coords/extra_coords/extracoords-*"]
    types = ["ndcube.extra_coords.extra_coords.ExtraCoords"]

    def from_yaml_tree(self, node, tag, ctx):
        from ndcube.extra_coords.extra_coords import ExtraCoords
        extra_coords = ExtraCoords()
        extra_coords._wcs = node.get("wcs")
        extra_coords._mapping = node.get("mapping")
        extra_coords._lookup_tables = node.get("lookup_tables", [])
        extra_coords._dropped_tables = node.get("dropped_tables")
        extra_coords._ndcube = node.get("ndcube")
        return extra_coords

    def to_yaml_tree(self, extracoords, tag, ctx):
        node = {}
        if extracoords._wcs is not None:
            node["wcs"] = extracoords._wcs
        if extracoords._mapping is not None:
            node["mapping"] = extracoords._mapping
        if extracoords._lookup_tables:
            node["lookup_tables"] = extracoords._lookup_tables
        if extracoords._dropped_tables is not None:
            node["dropped_tables"] = extracoords._dropped_tables
        node["ndcube"] = extracoords._ndcube
        return node



"""
Tests for roundtrip serialization of NDCube with various GWCS types.

TODO: Add tests for the roundtrip serialization of NDCube with ResampledLowLevelWCS, ReorderedLowLevelWCS, and CompoundLowLevelWCS when using astropy.wcs.WCS.
"""

import pytest

import asdf

from ndcube import NDCube
from ndcube.conftest import data_nd
from ndcube.tests.helpers import assert_cubes_equal
from ndcube.wcs.wrappers import CompoundLowLevelWCS, ReorderedLowLevelWCS, ResampledLowLevelWCS


@pytest.fixture
def create_ndcube_resampledwcs(gwcs_3d_lt_ln_l):
    shape = (2, 3, 4)
    new_wcs = ResampledLowLevelWCS(wcs=gwcs_3d_lt_ln_l, factor=2 , offset=1)
    data = data_nd(shape)
    return NDCube(data=data, wcs=new_wcs)


def test_serialization_resampled(create_ndcube_resampledwcs, tmp_path):
    ndc = create_ndcube_resampledwcs
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["ndcube"] = ndc
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        loaded_ndcube = af["ndcube"]

        loaded_resampledwcs = loaded_ndcube.wcs.low_level_wcs
        resampledwcs = ndc.wcs.low_level_wcs
        assert (loaded_resampledwcs._factor == resampledwcs._factor).all()
        assert (loaded_resampledwcs._offset == resampledwcs._offset).all()

        assert_cubes_equal(loaded_ndcube, ndc)


@pytest.fixture
def create_ndcube_reorderedwcs(gwcs_3d_lt_ln_l):
    shape = (2, 3, 4)
    new_wcs = ReorderedLowLevelWCS(wcs = gwcs_3d_lt_ln_l, pixel_order=[1, 2, 0] ,world_order=[2, 0, 1])
    data = data_nd(shape)
    return NDCube(data = data, wcs =new_wcs)


def test_serialization_reordered(create_ndcube_reorderedwcs, tmp_path):
    ndc = create_ndcube_reorderedwcs
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["ndcube"] = ndc
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        loaded_ndcube = af["ndcube"]

        loaded_reorderedwcs = loaded_ndcube.wcs.low_level_wcs
        reorderedwcs = ndc.wcs.low_level_wcs
        assert (loaded_reorderedwcs._pixel_order == reorderedwcs._pixel_order)
        assert (loaded_reorderedwcs._world_order == reorderedwcs._world_order)

        assert_cubes_equal(loaded_ndcube, ndc)

@pytest.fixture
def create_ndcube_compoundwcs(gwcs_2d_lt_ln, time_and_simple_extra_coords_2d):

    shape = (1, 2, 3, 4)
    new_wcs = CompoundLowLevelWCS(gwcs_2d_lt_ln, time_and_simple_extra_coords_2d.wcs, mapping = [0, 1, 2, 3])
    data = data_nd(shape)
    return NDCube(data = data, wcs = new_wcs)


def test_serialization_compoundwcs(create_ndcube_compoundwcs, tmp_path):
    ndc = create_ndcube_compoundwcs
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["ndcube"] = ndc
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        loaded_ndcube = af["ndcube"]
        assert_cubes_equal(loaded_ndcube, ndc)
        assert (loaded_ndcube.wcs.low_level_wcs.mapping.mapping == ndc.wcs.low_level_wcs.mapping.mapping)
        assert (loaded_ndcube.wcs.low_level_wcs.atol == ndc.wcs.low_level_wcs.atol)



"""
This file contains a set of common fixtures to get a set of different but
predictable NDCube objects.
"""
import logging

import numpy as np
import pytest
from gwcs import coordinate_frames as cf
from gwcs import wcs

import astropy.nddata
import astropy.units as u
from astropy import coordinates as coord
from astropy.coordinates import SkyCoord
from astropy.modeling import models
from astropy.nddata import StdDevUncertainty
from astropy.time import Time, TimeDelta
from astropy.wcs import WCS

from ndcube import ExtraCoords, GlobalCoords, NDCube, NDCubeSequence, NDMeta
from ndcube.extra_coords.table_coord import (
    QuantityTableCoordinate,
    SkyCoordTableCoordinate,
    TimeTableCoordinate,
)
from ndcube.tests import helpers

# Force MPL to use non-gui backends for testing.
try:
    import matplotlib
    import matplotlib.pyplot as plt
except ImportError:
    HAVE_MATPLOTLIB = False
else:
    HAVE_MATPLOTLIB = True
    matplotlib.use('Agg')


console_logger = logging.getLogger()
console_logger.setLevel('INFO')

################################################################################
# Helper Functions
################################################################################

def time_lut(shape):
    base_time = Time('2000-01-01', format='fits', scale='utc')
    return Time([base_time + TimeDelta(60 * i, format='sec') for i in range(shape[0])])

def skycoord_2d_lut(shape):
    total_len = np.prod(shape)
    data = (np.arange(total_len).reshape(shape),
            np.arange(total_len, total_len * 2).reshape(shape))
    return SkyCoord(*data, unit=u.deg)


def data_nd(shape, dtype=float):
    nelem = np.prod(shape)
    return np.arange(nelem, dtype=dtype).reshape(shape)


def time_extra_coords(shape, axis, base):
    return ExtraCoords.from_lookup_tables(
        ('time',),
        (axis,),
        (base + TimeDelta([i * 60 for i in range(shape[axis])], format='sec'),))


def gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l, time_axis, time_base, global_coords=None):
    shape = (10, 5, 8)
    wcs_3d_lt_ln_l.array_shape = shape
    data_cube = data_nd(shape)
    mask = data_cube < 0
    meta = {"message": "hello world"}
    unit = u.ph
    extra_coords = time_extra_coords(shape, time_axis, time_base)
    cube = NDCube(data_cube,
                  wcs_3d_lt_ln_l,
                  mask=mask,
                  uncertainty=data_cube,
                  meta=meta,
                  unit=unit)
    cube._extra_coords = extra_coords

    if global_coords:
        cube._global_coords = global_coords

    return cube


################################################################################
# WCS Fixtures
################################################################################

@pytest.fixture
def gwcs_4d_t_l_lt_ln():
    """
    Creates a 4D GWCS object with time, wavelength, and celestial coordinates.

    - Time: Axis 0
    - Wavelength: Axis 1
    - Sky: Axes 2 and 3

    Returns:
        wcs.WCS: 4D GWCS object.
    """

    time_model = models.Identity(1)
    time_frame = cf.TemporalFrame(axes_order=(0, ), unit=u.s,
                                  reference_frame=Time("2000-01-01T00:00:00"))

    wave_frame = cf.SpectralFrame(axes_order=(1, ), unit=u.m, axes_names=('wavelength',))
    wave_model = models.Scale(0.2)

    shift  = models.Shift(-5) & models.Shift(0)
    scale  = models.Scale(5) & models.Scale(20)
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(0, 0, 180)
    cel_model = shift | scale | tan | celestial_rotation
    sky_frame = cf.CelestialFrame(axes_order=(2, 3), name='icrs',
                                    reference_frame=coord.ICRS(),
                                    axes_names=("longitude", "latitude"))

    transform = time_model & wave_model & cel_model

    frame = cf.CompositeFrame([time_frame, wave_frame, sky_frame])
    detector_frame = cf.CoordinateFrame(name="detector", naxes=4,
                                        axes_order=(0, 1, 2, 3),
                                        axes_type=("pixel", "pixel", "pixel", "pixel"),
                                        unit=(u.pix, u.pix, u.pix, u.pix))

    return (wcs.WCS(forward_transform=transform, output_frame=frame, input_frame=detector_frame))

@pytest.fixture
def gwcs_3d_lt_ln_l():
    """
    Creates a 3D GWCS object with celestial coordinates and wavelength.

    - Sky: Axes 0 and 1
    - Wavelength: Axis 2

    Returns:
        wcs.WCS: 3D GWCS object.
    """

    shift  = models.Shift(-5) & models.Identity(1)
    scale  = models.Scale(5) & models.Scale(10)
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(0, 0, 180)
    cel_model = shift | scale | tan | celestial_rotation
    sky_frame = cf.CelestialFrame(axes_order=(0, 1), name='icrs',
                                    reference_frame=coord.ICRS(),
                                    axes_names=("longitude", "latitude"))

    wave_model = models.Identity(1) | models.Scale(0.2) | models.Shift(10)
    wave_frame = cf.SpectralFrame(axes_order=(2, ), unit=u.nm, axes_names=("wavelength",))

    transform = cel_model & wave_model

    frame = cf.CompositeFrame([sky_frame, wave_frame])
    detector_frame = cf.CoordinateFrame(name="detector", naxes=3,
                                        axes_order=(0, 1, 2),
                                        axes_type=("pixel", "pixel", "pixel"),
                                        axes_names=("x", "y", "z"), unit=(u.pix, u.pix, u.pix))

    return (wcs.WCS(forward_transform=transform, output_frame=frame, input_frame=detector_frame))

@pytest.fixture
def gwcs_3d_ln_lt_t_rotated():
    """
    Creates a 3D GWCS object with celestial coordinates and wavelength, including rotation.

    - Sky: Axes 0 and 1
    - Wavelength: Axis 2

    Returns:
        wcs.WCS: 3D GWCS object with rotation.
    """
    shift  = models.Shift(-5) & models.Identity(1)
    scale  = models.Scale(5) & models.Scale(10)
    matrix = np.array([[1.290551569736E-05, 5.9525007864732E-06],
                    [5.0226382102765E-06 , -1.2644844123757E-05]])
    rotation = models.AffineTransformation2D(matrix)
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(0, 0, 180)
    cel_model = shift | scale| rotation | tan | celestial_rotation
    sky_frame = cf.CelestialFrame(axes_order=(0, 1), name='icrs',
                                    reference_frame=coord.ICRS(),
                                    axes_names=("longitude", "latitude"))

    wave_model = models.Identity(1) | models.Scale(0.2) | models.Shift(10)
    wave_frame = cf.SpectralFrame(axes_order=(2, ), unit=u.nm, axes_names=("wavelength",))

    transform = cel_model & wave_model

    frame = cf.CompositeFrame([sky_frame, wave_frame])
    detector_frame = cf.CoordinateFrame(name="detector", naxes=3,
                                        axes_order=(0, 1, 2),
                                        axes_type=("pixel", "pixel", "pixel"),
                                        axes_names=("x", "y", "z"), unit=(u.pix, u.pix, u.pix))

    return (wcs.WCS(forward_transform=transform, output_frame=frame, input_frame=detector_frame))

@pytest.fixture
def gwcs_2d_lt_ln():
    """
    Creates a 2D GWCS object with celestial coordinates.

    - Sky: Axes 0 and 1

    Returns:
        wcs.WCS: 2D GWCS object.
    """
    shift  = models.Shift(-5) & models.Shift(-5)
    scale  = models.Scale(2) & models.Scale(4)
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(0, 0, 180)
    cel_model = shift | scale | tan | celestial_rotation
    input_frame = cf.Frame2D(name="detector", axes_names=("x", "y"))
    sky_frame = cf.CelestialFrame(axes_order=(0, 1), name='icrs',
                                    reference_frame=coord.ICRS(),
                                    axes_names=("longitude", "latitude"))

    return (wcs.WCS(forward_transform=cel_model, output_frame=sky_frame, input_frame=input_frame))


@pytest.fixture
def gwcs_2d_t_f_linear():
    """
    2D gWCS for a dynamic spectrum: uniform time (array axis 1 / X) and linear
    frequency (array axis 0 / Y).
    """
    time_model = models.Scale(14.0)
    freq_model = models.Scale(1e6)

    time_frame = cf.TemporalFrame(axes_order=(0,), unit=u.s,
                                  reference_frame=Time("2024-03-23T00:03:23"))
    freq_frame = cf.SpectralFrame(axes_order=(1,), unit=u.Hz, axes_names=('frequency',))

    transform = time_model & freq_model
    frame = cf.CompositeFrame([time_frame, freq_frame])
    detector_frame = cf.CoordinateFrame(name="detector", naxes=2,
                                        axes_order=(0, 1),
                                        axes_type=("pixel", "pixel"),
                                        unit=(u.pix, u.pix))
    return wcs.WCS(forward_transform=transform, output_frame=frame,
                   input_frame=detector_frame)


@pytest.fixture
def gwcs_2d_t_f_log():
    """
    2D gWCS for a dynamic spectrum: irregularly-spaced time (array axis 1 / X)
    and log-spaced frequency (array axis 0 / Y) via Tabular1D lookup tables.
    """
    times_s = np.array([0.0, 14.0, 27.4, 41.1, 55.2, 67.8, 82.3, 95.9, 109.1, 122.5])
    freqs_hz = np.logspace(np.log10(3.992e6), np.log10(978.572e6), 16)

    time_model = models.Tabular1D(points=np.arange(10), lookup_table=times_s,
                                  method='linear', bounds_error=False)
    freq_model = models.Tabular1D(points=np.arange(16), lookup_table=freqs_hz,
                                  method='linear', bounds_error=False)

    time_frame = cf.TemporalFrame(axes_order=(0,), unit=u.s,
                                  reference_frame=Time("2024-03-23T00:03:23"))
    freq_frame = cf.SpectralFrame(axes_order=(1,), unit=u.Hz, axes_names=('frequency',))

    transform = time_model & freq_model
    frame = cf.CompositeFrame([time_frame, freq_frame])
    detector_frame = cf.CoordinateFrame(name="detector", naxes=2,
                                        axes_order=(0, 1),
                                        axes_type=("pixel", "pixel"),
                                        unit=(u.pix, u.pix))
    return wcs.WCS(forward_transform=transform, output_frame=frame,
                   input_frame=detector_frame)

@pytest.fixture
def wcs_4d_t_l_lt_ln():
    header = {
        'CTYPE1': 'TIME    ',
        'CUNIT1': 'min',
        'CDELT1': 0.4,
        'CRPIX1': 0,
        'CRVAL1': 0,

        'CTYPE2': 'WAVE    ',
        'CUNIT2': 'Angstrom',
        'CDELT2': 0.2,
        'CRPIX2': 0,
        'CRVAL2': 0,

        'CTYPE3': 'HPLT-TAN',
        'CUNIT3': 'arcsec',
        'CDELT3': 20,
        'CRPIX3': 0,
        'CRVAL3': 0,

        'CTYPE4': 'HPLN-TAN',
        'CUNIT4': 'arcsec',
        'CDELT4': 5,
        'CRPIX4': 5,
        'CRVAL4': 0,

        'DATEREF': "2020-01-01T00:00:00"
    }
    return WCS(header=header)


@pytest.fixture
def wcs_4d_lt_t_l_ln():
    header = {
        'CTYPE1': 'HPLT-TAN',
        'CUNIT1': 'arcsec',
        'CDELT1': 20,
        'CRPIX1': 0,
        'CRVAL1': 0,

        'CTYPE2': 'TIME    ',
        'CUNIT2': 'min',
        'CDELT2': 0.4,
        'CRPIX2': 0,
        'CRVAL2': 0,

        'CTYPE3': 'WAVE    ',
        'CUNIT3': 'Angstrom',
        'CDELT3': 0.2,
        'CRPIX3': 0,
        'CRVAL3': 0,

        'CTYPE4': 'HPLN-TAN',
        'CUNIT4': 'arcsec',
        'CDELT4': 5,
        'CRPIX4': 5,
        'CRVAL4': 0,

        'DATEREF': "2020-01-01T00:00:00"
    }
    return WCS(header=header)


@pytest.fixture
def wcs_3d_l_lt_ln():
    header = {
        'CTYPE1': 'WAVE    ',
        'CUNIT1': 'Angstrom',
        'CDELT1': 0.2,
        'CRPIX1': 0,
        'CRVAL1': 10,

        'CTYPE2': 'HPLT-TAN',
        'CUNIT2': 'arcsec',
        'CDELT2': 5,
        'CRPIX2': 5,
        'CRVAL2': 0,

        'CTYPE3': 'HPLN-TAN',
        'CUNIT3': 'arcsec',
        'CDELT3': 10,
        'CRPIX3': 0,
        'CRVAL3': 0,
    }

    return WCS(header=header)


@pytest.fixture
def wcs_3d_lt_ln_l():
    header = {

        'CTYPE1': 'HPLN-TAN',
        'CUNIT1': 'arcsec',
        'CDELT1': 10,
        'CRPIX1': 0,
        'CRVAL1': 0,

        'CTYPE2': 'HPLT-TAN',
        'CUNIT2': 'arcsec',
        'CDELT2': 5,
        'CRPIX2': 5,
        'CRVAL2': 0,

        'CTYPE3': 'WAVE    ',
        'CUNIT3': 'Angstrom',
        'CDELT3': 0.2,
        'CRPIX3': 0,
        'CRVAL3': 10,
    }

    return WCS(header=header)


@pytest.fixture
def wcs_3d_wave_lt_ln():
    header = {
        'CTYPE1': 'WAVE    ',
        'CUNIT1': 'Angstrom',
        'CDELT1': 0.2,
        'CRPIX1': 0,
        'CRVAL1': 10,

        'CTYPE2': 'HPLT-TAN',
        'CUNIT2': 'deg',
        'CDELT2': 0.5,
        'CRPIX2': 2,
        'CRVAL2': 0.5,

        'CTYPE3': 'HPLN-TAN    ',
        'CUNIT3': 'deg',
        'CDELT3': 0.4,
        'CRPIX3': 2,
        'CRVAL3': 1,
    }
    return WCS(header=header)


@pytest.fixture
def wcs_2d_lt_ln():
    spatial = {
        'CTYPE1': 'HPLT-TAN',
        'CUNIT1': 'arcsec',
        'CDELT1': 2,
        'CRPIX1': 5,
        'CRVAL1': 0,

        'CTYPE2': 'HPLN-TAN',
        'CUNIT2': 'arcsec',
        'CDELT2': 4,
        'CRPIX2': 5,
        'CRVAL2': 0,
    }
    return WCS(header=spatial)


@pytest.fixture
def wcs_1d_l():
    spatial = {
        'CNAME1': 'spectral',
        'CTYPE1': 'WAVE',
        'CUNIT1': 'nm',
        'CDELT1': 0.5,
        'CRPIX1': 2,
        'CRVAL1': 0.5,
    }
    return WCS(header=spatial)


@pytest.fixture
def wcs_3d_ln_lt_t_rotated():
    h_rotated = {
        'CTYPE1': 'HPLN-TAN',
        'CUNIT1': 'arcsec',
        'CDELT1': 0.4,
        'CRPIX1': 0,
        'CRVAL1': 0,
        'NAXIS1': 5,

        'CTYPE2': 'HPLT-TAN',
        'CUNIT2': 'arcsec',
        'CDELT2': 0.5,
        'CRPIX2': 0,
        'CRVAL2': 0,
        'NAXIS2': 5,

        'CTYPE3': 'TIME    ',
        'CUNIT3': 's',
        'CDELT3': 3,
        'CRPIX3': 0,
        'CRVAL3': 0,
        'NAXIS3': 2,

        'DATEREF': "2020-01-01T00:00:00",

        'PC1_1': 0.714963912964,
        'PC1_2': -0.699137151241,
        'PC1_3': 0.0,
        'PC2_1': 0.699137151241,
        'PC2_2': 0.714963912964,
        'PC2_3': 0.0,
        'PC3_1': 0.0,
        'PC3_2': 0.0,
        'PC3_3': 1.0
    }
    return WCS(header=h_rotated)


@pytest.fixture
def wcs_3d_ln_lt_l_coupled():
    # WCS for a 3D data cube with two celestial axes and one wavelength axis.
    # The latitudinal dimension is coupled to the third pixel dimension through
    # a single off diagonal element in the PCij matrix
    header = {
        'CTYPE1': 'HPLN-TAN',
        'CRPIX1': 5,
        'CDELT1': 5,
        'CUNIT1': 'arcsec',
        'CRVAL1': 0.0,

        'CTYPE2': 'HPLT-TAN',
        'CRPIX2': 5,
        'CDELT2': 5,
        'CUNIT2': 'arcsec',
        'CRVAL2': 0.0,

        'CTYPE3': 'WAVE',
        'CRPIX3': 1.0,
        'CDELT3': 1,
        'CUNIT3': 'Angstrom',
        'CRVAL3': 1.0,

        'PC1_1': 1,
        'PC1_2': 0,
        'PC1_3': 0,
        'PC2_1': 0,
        'PC2_2': 1,
        'PC2_3': -1.0,
        'PC3_1': 0.0,
        'PC3_2': 0.0,
        'PC3_3': 1.0,

        'WCSAXES': 3,

        'DATEREF': "2020-01-01T00:00:00"
    }
    return WCS(header=header)


@pytest.fixture
def wcs_3d_ln_lt_t_coupled():
    # WCS for a 3D data cube with two celestial axes and one time axis.
    header = {
        'CTYPE1': 'HPLN-TAN',
        'CRPIX1': 5,
        'CDELT1': 5,
        'CUNIT1': 'arcsec',
        'CRVAL1': 0.0,

        'CTYPE2': 'HPLT-TAN',
        'CRPIX2': 5,
        'CDELT2': 5,
        'CUNIT2': 'arcsec',
        'CRVAL2': 0.0,

        'CTYPE3': 'UTC',
        'CRPIX3': 1.0,
        'CDELT3': 1,
        'CUNIT3': 's',
        'CRVAL3': 1.0,

        'PC1_1': 1,
        'PC1_2': 0,
        'PC1_3': 0,
        'PC2_1': 0,
        'PC2_2': 1,
        'PC2_3': 0,
        'PC3_1': 0,
        'PC3_2': 1,
        'PC3_3': 1,

        'WCSAXES': 3,

        'DATEREF': "2020-01-01T00:00:00"
    }
    return WCS(header=header)


################################################################################
# Extra and Global Coords Fixtures
################################################################################


@pytest.fixture
def simple_extra_coords_3d():
    return ExtraCoords.from_lookup_tables(('time', 'hello', 'bye'),
                                          (0, 1, 2),
                                          (list(range(2)) * u.pix,
                                           list(range(3)) * u.pix,
                                           list(range(4)) * u.pix
                                           )
                                          )


@pytest.fixture
def time_and_simple_extra_coords_2d():
    return ExtraCoords.from_lookup_tables(("time", "hello"),
                                          (0, 1),
                                          (Time(["2000-01-01T12:00:00", "2000-01-02T12:00:00"],
                                                scale="utc", format="fits"),
                                           list(range(3)) * u.pix)
                                          )


@pytest.fixture
def extra_coords_3d():
    coord0 = Time(["2000-01-01T12:00:00", "2000-01-02T12:00:00"], scale="utc", format="fits")
    coord1 = list(range(3)) * u.pix
    coord2 = list(range(4)) * u.m
    return ExtraCoords.from_lookup_tables(('time', 'bye', 'hello'),
                                          (0, 1, 2),
                                          (coord0, coord1, coord2)
                                          )


@pytest.fixture
def extra_coords_sharing_axis():
    return ExtraCoords.from_lookup_tables(('hello', 'bye'),
                                          (1, 1),
                                          (list(range(3)) * u.m,
                                           list(range(3)) * u.keV,
                                           )
                                          )

################################################################################
# NDCube Fixtures
# NOTE: If you add more fixtures please add to the all_ndcubes fixture
################################################################################

@pytest.fixture
def ndcube_gwcs_2d_t_f_linear(gwcs_2d_t_f_linear):
    shape = (16, 10)  # (n_freq, n_time): freq on Y axis, time on X axis
    gwcs_2d_t_f_linear.array_shape = shape
    return NDCube(data_nd(shape), wcs=gwcs_2d_t_f_linear)


@pytest.fixture
def ndcube_gwcs_2d_t_f_log(gwcs_2d_t_f_log):
    shape = (16, 10)  # (n_freq, n_time): freq on Y axis, time on X axis
    gwcs_2d_t_f_log.array_shape = shape
    return NDCube(data_nd(shape), wcs=gwcs_2d_t_f_log)


@pytest.fixture
def ndcube_gwcs_4d_ln_lt_l_t(gwcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    gwcs_4d_t_l_lt_ln.array_shape = shape
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=gwcs_4d_t_l_lt_ln)


@pytest.fixture
def ndcube_gwcs_4d_ln_lt_l_t_unit(gwcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    gwcs_4d_t_l_lt_ln.array_shape = shape
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=gwcs_4d_t_l_lt_ln, unit=u.DN)


@pytest.fixture
def ndcube_gwcs_3d_ln_lt_l(gwcs_3d_lt_ln_l):
    shape = (2, 3, 4)
    gwcs_3d_lt_ln_l.array_shape = shape
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=gwcs_3d_lt_ln_l)


@pytest.fixture
def ndcube_gwcs_3d_rotated(gwcs_3d_lt_ln_l, simple_extra_coords_3d):
    data_rotated = np.array([[[1, 2, 3, 4, 6], [2, 4, 5, 3, 1], [0, -1, 2, 4, 2], [3, 5, 1, 2, 0]],
                             [[2, 4, 5, 1, 3], [1, 5, 2, 2, 4], [2, 3, 4, 0, 5], [0, 1, 2, 3, 4]]])
    cube = NDCube(
        data_rotated,
        wcs=gwcs_3d_lt_ln_l)
    cube._extra_coords = simple_extra_coords_3d
    return cube


@pytest.fixture
def ndcube_gwcs_3d_ln_lt_l_ec_dropped_dim(gwcs_3d_lt_ln_l, time_and_simple_extra_coords_2d):
    shape = (2, 3, 4)
    gwcs_3d_lt_ln_l.array_shape = shape
    data_cube = data_nd(shape)
    cube =  NDCube(data_cube, wcs=gwcs_3d_lt_ln_l)
    cube._extra_coords = time_and_simple_extra_coords_2d[0]
    return cube


@pytest.fixture
def ndcube_gwcs_3d_ln_lt_l_ec_q_t_gc(gwcs_3d_lt_ln_l):
    shape = (3, 3, 4)
    gwcs_3d_lt_ln_l.array_shape = shape
    data_cube = data_nd(shape)
    cube =  NDCube(data_cube, wcs=gwcs_3d_lt_ln_l)
    coord1 = 1 * u.m
    cube.global_coords.add('name1', 'custom:physical_type1', coord1)
    cube.extra_coords.add("time", 0, time_lut(shape))
    cube.extra_coords.add("exposure_lut", 1, range(shape[1]) * u.s)
    return cube


@pytest.fixture
def ndcube_gwcs_2d_ln_lt_mask(gwcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape)
    mask = np.zeros(shape, dtype=bool)
    mask[1, 1] = True
    mask[2, 0] = True
    mask[3, 3] = True
    mask[4:6, :4] = True
    return NDCube(data_cube, wcs=gwcs_2d_lt_ln, mask=mask)


@pytest.fixture
def ndcube_4d_ln_l_t_lt(wcs_4d_lt_t_l_ln):
    shape = (5, 10, 12, 8)
    wcs_4d_lt_t_l_ln.array_shape = shape
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=wcs_4d_lt_t_l_ln)


@pytest.fixture
def ndcube_4d_ln_lt_l_t(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    wcs_4d_t_l_lt_ln.array_shape = shape
    data_cube = data_nd(shape, dtype=int)
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln)


@pytest.fixture
def ndcube_4d_axis_aware_meta(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    wcs_4d_t_l_lt_ln.array_shape = shape
    data_cube = data_nd(shape, dtype=int)
    meta = NDMeta({"a": "scalar",
                   "slit position": np.arange(shape[0], dtype=int),
                   "pixel label": np.arange(np.prod(shape[:2])).reshape(shape[:2]),
                   "line": ["Si IV"] * shape[2],
                   "exposure time": ([2] * shape[-1]) * u.s},
                  axes={"slit position": 0,
                        "pixel label": (0, 1),
                        "line": (2,),
                        "exposure time": 3})
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln, meta=meta)


@pytest.fixture
def ndcube_4d_uncertainty(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    data_cube = data_nd(shape)
    uncertainty = np.sqrt(data_cube)
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln, uncertainty=uncertainty)


@pytest.fixture
def ndcube_4d_mask(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    data_cube = data_nd(shape)
    uncertainty = np.sqrt(data_cube)
    mask = data_cube % 2
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln, uncertainty=uncertainty, mask=mask)


@pytest.fixture
def ndcube_4d_extra_coords(wcs_4d_t_l_lt_ln, simple_extra_coords_3d):
    shape = (5, 8, 10, 12)
    data_cube = data_nd(shape)
    cube = NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln)
    cube._extra_coords = simple_extra_coords_3d
    return cube


@pytest.fixture
def ndcube_4d_unit_uncertainty(wcs_4d_t_l_lt_ln):
    shape = (5, 8, 10, 12)
    data_cube = data_nd(shape)
    uncertainty = np.sqrt(data_cube)
    return NDCube(data_cube, wcs=wcs_4d_t_l_lt_ln,
                  unit=u.J, uncertainty=uncertainty)


@pytest.fixture
def ndcube_4d(request):
    """
    This is a meta fixture for parametrizing all the 4D ndcubes.
    """
    return request.getfixturevalue("ndcube_4d_" + request.param)


@pytest.fixture
def ndcube_3d_ln_lt_l(wcs_3d_l_lt_ln, simple_extra_coords_3d):
    shape = (2, 3, 4)
    wcs_3d_l_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_l_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    cube._extra_coords = simple_extra_coords_3d
    cube._extra_coords._ndcube = cube
    return cube


@pytest.fixture
def ndcube_3d_ln_lt_l_ec_all_axes(wcs_3d_l_lt_ln, extra_coords_3d):
    shape = (2, 3, 4)
    wcs_3d_l_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_l_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    cube._extra_coords = extra_coords_3d
    cube._extra_coords._ndcube = cube
    return cube


@pytest.fixture
def ndcube_3d_ln_lt_l_ec_sharing_axis(wcs_3d_l_lt_ln, extra_coords_sharing_axis):
    shape = (2, 3, 4)
    wcs_3d_l_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_l_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    cube._extra_coords = extra_coords_sharing_axis
    cube._extra_coords._ndcube = cube
    return cube


@pytest.fixture
def ndcube_3d_ln_lt_l_ec_time(wcs_3d_l_lt_ln, time_and_simple_extra_coords_2d):
    shape = (2, 3, 4)
    wcs_3d_l_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_l_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    cube._extra_coords = time_and_simple_extra_coords_2d
    cube._extra_coords._ndcube = cube
    return cube


@pytest.fixture
def ndcube_3d_wave_lt_ln_ec_time(wcs_3d_wave_lt_ln):
    shape = (3, 4, 5)
    wcs_3d_wave_lt_ln.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    cube = NDCube(
        data,
        wcs_3d_wave_lt_ln,
        mask=mask,
        uncertainty=data,
    )
    base_time = Time('2000-01-01', format='fits', scale='utc')
    timestamps = Time([base_time + TimeDelta(60 * i, format='sec') for i in range(data.shape[0])])
    cube.extra_coords.add('time', 0, timestamps)
    return cube


@pytest.fixture
def ndcube_3d_rotated(wcs_3d_ln_lt_t_rotated, simple_extra_coords_3d):
    data_rotated = np.array([[[1, 2, 3, 4, 6], [2, 4, 5, 3, 1], [0, -1, 2, 4, 2], [3, 5, 1, 2, 0]],
                             [[2, 4, 5, 1, 3], [1, 5, 2, 2, 4], [2, 3, 4, 0, 5], [0, 1, 2, 3, 4]]])
    mask_rotated = data_rotated >= 0
    cube = NDCube(
        data_rotated,
        wcs_3d_ln_lt_t_rotated,
        mask=mask_rotated,
        uncertainty=data_rotated,
    )
    cube._extra_coords = simple_extra_coords_3d
    return cube


@pytest.fixture
def ndcube_3d_coupled(wcs_3d_ln_lt_l_coupled):
    shape = (128, 256, 512)
    wcs_3d_ln_lt_l_coupled.array_shape = shape
    data = data_nd(shape)
    mask = data > 0
    return NDCube(
        data,
        wcs_3d_ln_lt_l_coupled,
        mask=mask,
        uncertainty=data,
    )


@pytest.fixture
def ndcube_3d_coupled_time(wcs_3d_ln_lt_t_coupled):
    shape = (128, 256, 512)
    wcs_3d_ln_lt_t_coupled.array_shape = shape
    data = data_nd(shape)
    return NDCube(
        data,
        wcs_3d_ln_lt_t_coupled,
    )


@pytest.fixture
def ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l):
    return gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l,
                                        1,
                                        Time('2000-01-01', format='fits', scale='utc'))


@pytest.fixture
def ndcube_2d_ln_lt(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln)


@pytest.fixture
def ndcube_2d_ln_lt_psf(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, psf=np.zeros(shape))


@pytest.fixture
def ndcube_2d_ln_lt_uncert(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape)
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty)


@pytest.fixture
def ndcube_2d_ln_lt_mask_uncert(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape)
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    mask = np.zeros(shape, dtype=bool)
    mask[1, 1] = True
    mask[2, 0] = True
    mask[3, 3] = True
    mask[4:6, :4] = True
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, mask=mask)


@pytest.fixture
def ndcube_2d_ln_lt_mask_uncert_unit_mask_false(wcs_2d_lt_ln):
    shape = (2, 3)
    unit = u.ct
    data_cube = data_nd(shape)
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    mask = False
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, mask=mask, unit=unit)


@pytest.fixture
def ndcube_2d_ln_lt_mask_uncert_unit_one_maskele_true(wcs_2d_lt_ln):
    shape = (2, 3)
    unit = u.ct
    data_cube = data_nd(shape)
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    mask = np.zeros(shape, dtype=bool)
    mask[0:1, 0] = True
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, mask=mask, unit=unit)


@pytest.fixture
def ndcube_2d_ln_lt_mask_uncert_unit_one_maskele_true_expected_unmask_false(wcs_2d_lt_ln):
    shape = (2, 3)
    unit = u.ct
    data_cube = np.array([[1.0, 1.0, 2.0],
                          [3.0, 4.0, 5.0]])
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    mask = np.zeros(shape, dtype=bool)
    mask[0:1, 0] = True
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, mask=mask, unit=unit)


@pytest.fixture
def ndcube_2d_ln_lt_mask_uncert_unit_one_maskele_true_expected_unmask_true(wcs_2d_lt_ln):
    unit = u.ct
    data_cube = np.array([[1.0, 1.0, 2.0],
                          [3.0, 4.0, 5.0]])
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    mask = False
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, mask=mask, unit=unit)


@pytest.fixture
def ndcube_2d_ln_lt_mask_uncert_unit_mask_true(wcs_2d_lt_ln):
    shape = (2, 3)
    unit = u.ct
    data_cube = data_nd(shape)
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    mask = True
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, mask=mask, unit=unit)


@pytest.fixture
def ndcube_2d_ln_lt_mask_uncert_unit_mask_true_expected_unmask_true(wcs_2d_lt_ln):
    unit = u.ct
    data_cube = np.array([[1.0, 1.0, 1.0],
                          [1.0, 1.0, 1.0]])
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    mask = False
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, mask=mask, unit=unit)


@pytest.fixture
def ndcube_2d_ln_lt_mask_uncert_unit_mask_true_expected_unmask_false(wcs_2d_lt_ln):
    unit = u.ct
    data_cube = np.array([[1.0, 1.0, 1.0],
                          [1.0, 1.0, 1.0]])
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    mask = True
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, mask=mask, unit=unit)


@pytest.fixture
def ndcube_2d_ln_lt_uncert_ec(wcs_2d_lt_ln):
    shape = (4, 9)
    data_cube = data_nd(shape)
    uncertainty = astropy.nddata.StdDevUncertainty(data_cube * 0.1)
    cube = NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty)
    cube.extra_coords.add(
        "time", 0,
        Time("2000-01-01 00:00", scale="utc") + TimeDelta(np.arange(shape[0])*60, format="sec"))
    return cube


@pytest.fixture
def ndcube_2d_ln_lt_units(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape).astype(float)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, unit=u.ct)


@pytest.fixture
def ndcube_2d_ln_lt_no_unit_no_unc(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape).astype(float)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln)


@pytest.fixture
def ndcube_2d_unit_unc(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape).astype(float)
    uncertainty = StdDevUncertainty(np.ones(shape)*0.2, unit=u.ct)

    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty, unit=u.ct)


@pytest.fixture
def ndcube_2d_uncertainty_no_unit(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape).astype(float)
    uncertainty = StdDevUncertainty(np.ones(shape)*0.2)

    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty)


@pytest.fixture
def ndcube_2d_ln_lt_mask(wcs_2d_lt_ln):
    shape = (10, 12)
    data_cube = data_nd(shape).astype(float)
    mask = np.ones(data_cube.shape, dtype=bool)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, mask=mask)


@pytest.fixture
def ndcube_2d_ln_lt_mask2(wcs_2d_lt_ln):
    shape = (2, 3)
    data_cube = data_nd(shape).astype(float)
    mask = np.ones(shape, dtype=bool)
    mask[0:1, 0] = False
    uncertainty=StdDevUncertainty(np.ones((2, 3)) * 0.05)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, mask=mask, uncertainty=uncertainty)


@pytest.fixture
def ndcube_2d_ln_lt_nomask(wcs_2d_lt_ln):
    shape = (2, 3)
    data_cube = data_nd(shape).astype(float)
    uncertainty=StdDevUncertainty(np.ones((2, 3)) * 0.05)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, uncertainty=uncertainty)


@pytest.fixture
def ndcube_2d_ln_lt_no_unit_no_unc_no_mask_2(wcs_2d_lt_ln):
    shape = (2, 3)
    data_cube = data_nd(shape).astype(float)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln)


@pytest.fixture
def ndcube_2d_ln_lt_unit_unc_mask(wcs_2d_lt_ln):
    shape = (2, 3)
    data_cube = data_nd(shape).astype(float)
    mask = np.ones(shape, dtype=bool)
    mask[:, 0] = False
    uncertainty=StdDevUncertainty(data_cube * 0.05)
    return NDCube(data_cube, wcs=wcs_2d_lt_ln, mask=mask, uncertainty=uncertainty, unit=u.ct)


@pytest.fixture
def ndcube_2d_dask(wcs_2d_lt_ln):
    da = pytest.importorskip("dask.array")
    shape = (8, 4)
    chunks = 2
    data = data_nd(shape).astype(float)
    darr = da.asarray(data, chunks=chunks)
    mask = np.zeros(shape, dtype=bool)
    da_mask = da.asarray(mask, chunks=chunks)
    uncert = data * 0.1
    da_uncert = StdDevUncertainty(da.asarray(uncert, chunks=chunks))
    return NDCube(darr, wcs=wcs_2d_lt_ln, uncertainty=da_uncert, mask=da_mask, unit=u.J)


@pytest.fixture
def nddata_2d_dask(ndcube_2d_dask):
    return ndcube_2d_dask.to_nddata(wcs=None)


@pytest.fixture
def ndcube_2d(request):
    """
    This is a meta fixture for parametrizing all the 2D ndcubes.
    """
    return request.getfixturevalue("ndcube_2d_" + request.param)


@pytest.fixture
def ndcube_1d_l(wcs_1d_l):
    shape = (10,)
    data_cube = data_nd(shape)
    return NDCube(data_cube, wcs=wcs_1d_l,
                  uncertainty=StdDevUncertainty(data_cube*0.1), unit=u.J)


@pytest.fixture(params=[
    "ndcube_gwcs_2d_t_f_linear",
    "ndcube_gwcs_2d_t_f_log",
    "ndcube_gwcs_4d_ln_lt_l_t",
    "ndcube_gwcs_4d_ln_lt_l_t_unit",
    "ndcube_gwcs_3d_ln_lt_l",
    "ndcube_gwcs_3d_rotated",
    "ndcube_gwcs_3d_ln_lt_l_ec_dropped_dim",
    "ndcube_gwcs_3d_ln_lt_l_ec_q_t_gc",
    "ndcube_gwcs_2d_ln_lt_mask",
    "ndcube_4d_ln_l_t_lt",
    "ndcube_4d_ln_lt_l_t",
    "ndcube_4d_axis_aware_meta",
    "ndcube_4d_uncertainty",
    "ndcube_4d_mask",
    "ndcube_4d_extra_coords",
    "ndcube_4d_unit_uncertainty",
    "ndcube_3d_ln_lt_l",
    "ndcube_3d_ln_lt_l_ec_all_axes",
    "ndcube_3d_ln_lt_l_ec_sharing_axis",
    "ndcube_3d_ln_lt_l_ec_time",
    "ndcube_3d_wave_lt_ln_ec_time",
    "ndcube_3d_rotated",
    "ndcube_3d_coupled",
    "ndcube_3d_coupled_time",
    "ndcube_3d_l_ln_lt_ectime",
    "ndcube_2d_ln_lt",
    "ndcube_2d_ln_lt_psf",
    "ndcube_2d_ln_lt_uncert",
    "ndcube_2d_ln_lt_mask_uncert",
    "ndcube_2d_ln_lt_mask_uncert_unit_mask_false",
    "ndcube_2d_ln_lt_mask_uncert_unit_one_maskele_true",
    "ndcube_2d_ln_lt_mask_uncert_unit_one_maskele_true_expected_unmask_false",
    "ndcube_2d_ln_lt_mask_uncert_unit_one_maskele_true_expected_unmask_true",
    "ndcube_2d_ln_lt_mask_uncert_unit_mask_true",
    "ndcube_2d_ln_lt_mask_uncert_unit_mask_true_expected_unmask_true",
    "ndcube_2d_ln_lt_mask_uncert_unit_mask_true_expected_unmask_false",
    "ndcube_2d_ln_lt_uncert_ec",
    "ndcube_2d_ln_lt_units",
    "ndcube_2d_ln_lt_no_unit_no_unc",
    "ndcube_2d_unit_unc",
    "ndcube_2d_uncertainty_no_unit",
    "ndcube_2d_ln_lt_mask",
    "ndcube_2d_ln_lt_mask2",
    "ndcube_2d_ln_lt_nomask",
    "ndcube_2d_dask",
    "ndcube_1d_l",
])
def all_ndcubes_names(request):
    return request.param

@pytest.fixture
def all_ndcubes(request, all_ndcubes_names):
    """
    All the above ndcube fixtures in order.
    """
    return request.getfixturevalue(all_ndcubes_names)


@pytest.fixture
def ndc(request):
    """
    A fixture for use with indirect to lookup other fixtures.
    """
    return request.getfixturevalue(request.param)


@pytest.fixture
def expected_cube(request):
    """
    A fixture for use with indirect to lookup other fixtures.
    """
    return request.getfixturevalue(request.param)


################################################################################
# NDCubeSequence Fixtures
################################################################################


@pytest.fixture
def ndcubesequence_4c_ln_lt_l(ndcube_3d_ln_lt_l):
    cube1 = ndcube_3d_ln_lt_l
    cube2 = ndcube_3d_ln_lt_l
    cube3 = ndcube_3d_ln_lt_l
    cube4 = ndcube_3d_ln_lt_l
    cube2.data[:] *= 2
    cube3.data[:] *= 3
    cube4.data[:] *= 4
    return NDCubeSequence([cube1, cube2, cube3, cube4])


@pytest.fixture
def ndcubesequence_4c_ln_lt_l_cax1(ndcube_3d_ln_lt_l):
    cube1 = ndcube_3d_ln_lt_l
    cube2 = ndcube_3d_ln_lt_l
    cube3 = ndcube_3d_ln_lt_l
    cube4 = ndcube_3d_ln_lt_l
    cube2.data[:] *= 2
    cube3.data[:] *= 3
    cube4.data[:] *= 4
    meta = helpers.ndmeta_et0_pr02((4, 2, 3, 4))
    return NDCubeSequence([cube1, cube2, cube3, cube4], common_axis=1, meta=meta)


@pytest.fixture
def ndcubesequence_3c_l_ln_lt_cax1(wcs_3d_lt_ln_l):
    common_axis = 1

    base_time1 = Time('2000-01-01', format='fits', scale='utc')
    gc1 = GlobalCoords()
    gc1.add('distance', 'custom:distance', 1*u.m)
    cube1 = gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l, 1, base_time1, gc1)

    shape = cube1.data.shape
    base_time2 = base_time1 + TimeDelta([shape[common_axis] * 60], format='sec')
    gc2 = GlobalCoords()
    gc2.add('distance', 'custom:distance', 2*u.m)
    gc2.add('global coord', 'custom:physical_type', 0*u.pix)
    cube2 = gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l, 1, base_time2, gc2)
    cube2.data[:] *= 2

    base_time3 = base_time2 + TimeDelta([shape[common_axis] * 60], format='sec')
    gc3 = GlobalCoords()
    gc3.add('distance', 'custom:distance', 3*u.m)
    cube3 = gen_ndcube_3d_l_ln_lt_ectime(wcs_3d_lt_ln_l, 1, base_time3, gc3)
    cube3.data[:] *= 3

    return NDCubeSequence([cube1, cube2, cube3], common_axis=common_axis)

################################################################################
# Table Coordinates
################################################################################


@pytest.fixture
def lut_1d_distance():
    lookup_table = u.Quantity(np.arange(10) * u.km)
    return QuantityTableCoordinate(lookup_table, names='x')


@pytest.fixture
def lut_3d_distance_mesh():
    lookup_table = (u.Quantity(np.arange(10) * u.km),
                    u.Quantity(np.arange(10, 20) * u.km),
                    u.Quantity(np.arange(20, 30) * u.km))

    return QuantityTableCoordinate(*lookup_table, names=['x', 'y', 'z'])


@pytest.fixture
def lut_1d_skycoord_no_mesh():
    sc = SkyCoord(range(10), range(10), unit=u.deg)
    return SkyCoordTableCoordinate(sc, mesh=False, names=['lon', 'lat'])


@pytest.fixture
def lut_2d_skycoord_no_mesh():
    data = np.arange(9).reshape(3, 3), np.arange(9, 18).reshape(3, 3)
    sc = SkyCoord(*data, unit=u.deg)
    return SkyCoordTableCoordinate(sc, mesh=False)


@pytest.fixture
def lut_2d_skycoord_mesh():
    sc = SkyCoord(range(10), range(10), unit=u.deg)
    return SkyCoordTableCoordinate(sc, mesh=True)


@pytest.fixture
def lut_3d_skycoord_mesh():
    sc = SkyCoord(range(10), range(10), range(10), unit=(u.deg, u.deg, u.AU))
    return SkyCoordTableCoordinate(sc, mesh=True)


@pytest.fixture
def lut_1d_time():
    data = Time(["2011-01-01T00:00:00",
                 "2011-01-01T00:00:10",
                 "2011-01-01T00:00:20",
                 "2011-01-01T00:00:30"], format="isot")
    return TimeTableCoordinate(data, names='time', physical_types='time')


@pytest.fixture
def lut_1d_wave():
    # TODO: Make this into a SpectralCoord object
    return QuantityTableCoordinate(range(10) * u.nm)


def pytest_runtest_teardown(item):
    # Clear the pyplot figure stack if it is not empty after the test
    # You can see these log messages by passing "-o log_cli=true" to pytest on the command line
    if HAVE_MATPLOTLIB and plt.get_fignums():
        console_logger.info(f"Removing {len(plt.get_fignums())} pyplot figure(s) "
                            f"left open by {item.name}")
        plt.close('all')



from asdf.extension import Converter


class TimeTableCoordConverter(Converter):
    tags = ["tag:sunpy.org:ndcube/extra_coords/table_coord/timetablecoordinate-*"]
    types = ["ndcube.extra_coords.table_coord.TimeTableCoordinate"]

    def from_yaml_tree(self, node, tag, ctx):
        from ndcube.extra_coords.table_coord import TimeTableCoordinate

        names = node.get("names")
        physical_types = node.get("physical_types")
        reference_time = node["reference_time"]
        return TimeTableCoordinate(
            node["table"],
            names=names,
            physical_types=physical_types,
            reference_time=reference_time,
        )

    def to_yaml_tree(self, timetablecoordinate, tag, ctx):
        node = {}
        node["table"] = timetablecoordinate.table
        if timetablecoordinate.names:
            node["names"] = timetablecoordinate.names
        if timetablecoordinate.physical_types is not None:
            node["physical_types"] = timetablecoordinate.physical_types
        node["reference_time"] = timetablecoordinate.reference_time

        return node


class QuantityTableCoordinateConverter(Converter):
    tags = ["tag:sunpy.org:ndcube/extra_coords/table_coord/quantitytablecoordinate-*"]
    types = ["ndcube.extra_coords.table_coord.QuantityTableCoordinate"]

    def from_yaml_tree(self, node, tag, ctx):
        from ndcube.extra_coords.table_coord import QuantityTableCoordinate

        names = node.get("names")
        mesh = node.get("mesh")
        physical_types = node.get("physical_types")
        quantitytablecoordinate = QuantityTableCoordinate(*node["table"], names=names, physical_types=physical_types)
        quantitytablecoordinate.unit = node["unit"]
        quantitytablecoordinate.mesh = mesh
        return quantitytablecoordinate

    def to_yaml_tree(self, quantitytablecoordinate, tag, ctx):
        node = {}
        node["unit"] = quantitytablecoordinate.unit
        node["table"] = quantitytablecoordinate.table
        if quantitytablecoordinate.names:
            node["names"] = quantitytablecoordinate.names
        node["mesh"] = quantitytablecoordinate.mesh
        if quantitytablecoordinate.physical_types is not None:
            node["physical_types"] = quantitytablecoordinate.physical_types

        return node


class SkyCoordTableCoordinateConverter(Converter):
    tags = ["tag:sunpy.org:ndcube/extra_coords/table_coord/skycoordtablecoordinate-*"]
    types = ["ndcube.extra_coords.table_coord.SkyCoordTableCoordinate"]

    def from_yaml_tree(self, node, tag, ctx):
        from ndcube.extra_coords.table_coord import SkyCoordTableCoordinate

        names = node.get("names")
        mesh = node.get("mesh")
        physical_types = node.get("physical_types")
        return SkyCoordTableCoordinate(node["table"], mesh=mesh, names=names, physical_types=physical_types)

    def to_yaml_tree(self, skycoordinatetablecoordinate, tag, ctx):
        node = {}
        node["table"] = skycoordinatetablecoordinate.table
        if skycoordinatetablecoordinate.names:
            node["names"] = skycoordinatetablecoordinate.names
        node["mesh"] = skycoordinatetablecoordinate.mesh
        if skycoordinatetablecoordinate.physical_types is not None:
            node["physical_types"] = skycoordinatetablecoordinate.physical_types

        return node


class MultipleTableCoordinateConverter(Converter):
    tags = ["tag:sunpy.org:ndcube/extra_coords/table_coord/multipletablecoordinate-*"]
    types = ["ndcube.extra_coords.table_coord.MultipleTableCoordinate"]

    def from_yaml_tree(self, node, tag, ctx):
        from ndcube.extra_coords.table_coord import MultipleTableCoordinate

        mtc = MultipleTableCoordinate(*node["table_coords"])
        mtc._dropped_coords = node["dropped_coords"]
        return mtc

    def to_yaml_tree(self, multipletablecoordinate, tag, ctx):
        node = {}
        node["table_coords"] = multipletablecoordinate._table_coords
        node["dropped_coords"] = multipletablecoordinate._dropped_coords
        return node




import numpy as np

from astropy.wcs.wcsapi.wrappers.base import BaseWCSWrapper

__all__ = ['ResampledLowLevelWCS']


class ResampledLowLevelWCS(BaseWCSWrapper):
    """
    A wrapper for a low-level WCS object that has down- or
    up-sampled pixel axes.

    Parameters
    ----------
    wcs : `~astropy.wcs.wcsapi.BaseLowLevelWCS`
        The original WCS for which to reorder axes

    factor : `int` or `float` or iterable of the same
        The factor by which to increase the pixel size for each pixel
        axis. If a scalar, the same factor is used for all axes.

    offset: `int` or `float` or iterable of the same
        The shift of the lower edge of the 0th pixel (i.e. the pixel
        coordinate -0.5) of the resampled grid relative to the lower
        edge of the 0th pixel in the original underlying pixel grid,
        in units of original pixel widths.  (See the schematic in the
        Notes section for a graphical example.) If a scalar, the grid
        will be shifted by the same amount in all dimensions.

    Notes
    -----
    Below is a schematic of how ResampledLowLevelWCS works. The asterisks show the
    corners of pixels in a grid before resampling, while the dashes and
    pipes show the edges of the resampled pixels. The resampling along the
    x-axis has been performed using a factor of 2 and offset of 1, respectively,
    while the resampling of the y-axis uses a factor of 3 and offset of 2.
    The right column and upper row of numbers along the side and bottom of the
    grids denote the edges and centres of the original pixel grid in the original
    pixel coordinates.  The left column and lower row gives the same locations in
    the pixel coordinates of the resampled grid.  Note that the resampled pixels
    have an (x, y) shape of (2, 3) relative to the original pixel grid.
    Also note, the left/lower edge of the 0th pixel in the resampled grid (i.e. pixel
    coord -0.5) is shifted relative to the left/lower edge of the original 0th pixel,
    and that shift is given by the offset (+1 in the x-axis and +2 along the y-axis),
    which is in units of original pixel widths.

    ::

        resampled  original
        factor=3
        offset=2

          0.5      4.5 *-----------*-----------*-----------*-----------*
                                   |                       |
          2/6       4              |                       |
                                   |                       |
          1/6      3.5 *           *           *           *           *
                                   |                       |
           0        3              |                       |
                                   |                       |
         -1/3      2.5 *           *           *           *           *
                                   |                       |
         -2/6       2              |                       |
                                   |                       |
         -0.5      1.5 *-----------*-----------*-----------*-----------*
                                   |                       |
         -4/6       1              |                       |
                                   |                       |
         -5/6      0.5 *           *           *           *           *
                                   |                       |
          -1        0              |                       |
                                   |                       |
         -1-1/6   -0.5 *           *           *           *           *
                     -0.5    0    0.5    1    1.5    2    2.5    3    3.5  original pixel indices
                      -1   -0.75 -0.5  -0.25   0    0.25  0.5   0.75   1   resampled pixel indices: factor=2, offset=1
    """

    def __init__(self, wcs, factor, offset=0):
        self._wcs = wcs
        if np.isscalar(factor):
            factor = [factor] * self.pixel_n_dim
        self._factor = np.array(factor)
        if len(self._factor) != self.pixel_n_dim:
            raise ValueError(f"Length of factor must equal number of dimensions {self.pixel_n_dim}.")
        if np.isscalar(offset):
            offset = [offset] * self.pixel_n_dim
        self._offset = np.array(offset)
        if len(self._offset) != self.pixel_n_dim:
            raise ValueError(f"Length of offset must equal number of dimensions {self.pixel_n_dim}.")

    def _top_to_underlying_pixels(self, top_pixels):
        # Convert user-facing pixel indices to the pixel grid of underlying WCS.
        factor = self._pad_dims(self._factor, top_pixels.ndim)
        offset = self._pad_dims(self._offset, top_pixels.ndim)
        return (top_pixels + 0.5) * factor - 0.5 + offset

    def _underlying_to_top_pixels(self, underlying_pixels):
        # Convert pixel indices of underlying pixel grid to user-facing grid.
        factor = self._pad_dims(self._factor, underlying_pixels.ndim)
        offset = self._pad_dims(self._offset, underlying_pixels.ndim)
        return (underlying_pixels + 0.5 - offset) / factor - 0.5

    def _pad_dims(self, arr, ndim):
        # Pad array with trailing degenerate dimensions.
        # This make scaling with pixel arrays easier.
        shape = np.ones(ndim, dtype=int)
        shape[0] = len(arr)
        return arr.reshape(tuple(shape))

    def pixel_to_world_values(self, *pixel_arrays):
        underlying_pixel_arrays = self._top_to_underlying_pixels(np.asarray(pixel_arrays))
        return self._wcs.pixel_to_world_values(*underlying_pixel_arrays)

    def world_to_pixel_values(self, *world_arrays):
        underlying_pixel_arrays = self._wcs.world_to_pixel_values(*world_arrays)
        top_pixel_arrays = self._underlying_to_top_pixels(np.asarray(underlying_pixel_arrays))
        return tuple(array for array in top_pixel_arrays)

    @property
    def pixel_shape(self):
        # Return pixel shape of resampled grid.
        # Where shape is an integer, return an int type as its required for some uses.
        if self._wcs.pixel_shape is None:
            return self._wcs.pixel_shape
        underlying_shape = np.asarray(self._wcs.pixel_shape)
        int_elements = np.isclose(np.mod(underlying_shape, self._factor), 0,
                                  atol=np.finfo(float).resolution)
        pixel_shape = underlying_shape / self._factor
        return tuple(int(np.rint(i)) if is_int else i
                     for i, is_int in zip(pixel_shape, int_elements))

    @property
    def pixel_bounds(self):
        if self._wcs.pixel_bounds is None:
            return self._wcs.pixel_bounds
        top_level_bounds = self._underlying_to_top_pixels(np.asarray(self._wcs.pixel_bounds))
        return [tuple(bounds) for bounds in top_level_bounds]



import pytest

import asdf

from ndcube.ndcollection import NDCollection
from ndcube.ndcube_sequence import NDCubeSequence
from ndcube.tests.helpers import assert_collections_equal


@pytest.fixture
def create_ndcollection_cube(
    ndcube_gwcs_3d_ln_lt_l,
    ndcube_gwcs_3d_ln_lt_l_ec_q_t_gc,
    ndcube_gwcs_3d_ln_lt_l_ec_dropped_dim,
):
    aligned_axes = ((1, 2), (1, 2), (1, 2))
    return NDCollection(
        [
            ("cube0", ndcube_gwcs_3d_ln_lt_l),
            ("cube1", ndcube_gwcs_3d_ln_lt_l_ec_q_t_gc),
            ("cube2", ndcube_gwcs_3d_ln_lt_l_ec_dropped_dim),
        ],
        aligned_axes=aligned_axes,
    )


def test_serialization_cube(create_ndcollection_cube, tmp_path):
    ndcollection = create_ndcollection_cube
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["ndcube_gwcs"] = ndcollection
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_collections_equal(af["ndcube_gwcs"], ndcollection)


@pytest.fixture(params=[{"meta": {"hello": "world"}}, {}])
def create_ndcollection_sequence(ndcube_gwcs_3d_ln_lt_l, ndcube_gwcs_3d_ln_lt_l_ec_dropped_dim, request):
    sequence02 = NDCubeSequence([ndcube_gwcs_3d_ln_lt_l, ndcube_gwcs_3d_ln_lt_l_ec_dropped_dim])
    sequence20 = NDCubeSequence([ndcube_gwcs_3d_ln_lt_l_ec_dropped_dim, ndcube_gwcs_3d_ln_lt_l])
    return NDCollection([("seq0", sequence02), ("seq1", sequence20)], aligned_axes="all", **request.param)


def test_serialization_sequence(create_ndcollection_sequence, tmp_path):
    ndcollection = create_ndcollection_sequence
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["ndcube_gwcs"] = ndcollection
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_collections_equal(af["ndcube_gwcs"], ndcollection)



import re
import copy

import numpy as np
import pytest

import astropy.units as u
import astropy.wcs
from astropy.coordinates import SpectralCoord
from astropy.nddata import UnknownUncertainty
from astropy.time import Time

from ndcube import ExtraCoords
from ndcube.tests import helpers
from ndcube.utils.exceptions import NDCubeUserWarning


def test_reproject_interpolation(ndcube_4d_ln_l_t_lt, wcs_4d_lt_t_l_ln):
    target_wcs_header = wcs_4d_lt_t_l_ln.low_level_wcs.to_header()
    target_wcs_header['CDELT3'] = 0.1   # original value = 0.2
    target_wcs = astropy.wcs.WCS(header=target_wcs_header)
    shape_out = (5, 20, 12, 8)

    resampled_cube = ndcube_4d_ln_l_t_lt.reproject_to(target_wcs, shape_out=shape_out)

    assert ndcube_4d_ln_l_t_lt.data.shape == (5, 10, 12, 8)
    assert resampled_cube.data.shape == (5, 20, 12, 8)


def test_reproject_invalid_wcs(ndcube_4d_ln_l_t_lt, wcs_3d_lt_ln_l):
    shape_out = (5, 20, 12, 8)

    with pytest.raises(Exception):
        _ = ndcube_4d_ln_l_t_lt.reproject_to(wcs_3d_lt_ln_l, shape_out=shape_out)


def test_reproject_with_header(ndcube_4d_ln_l_t_lt, wcs_4d_lt_t_l_ln):
    target_wcs_header = wcs_4d_lt_t_l_ln.low_level_wcs.to_header()
    shape_out = (5, 20, 12, 8)

    _ = ndcube_4d_ln_l_t_lt.reproject_to(target_wcs_header, shape_out=shape_out)


def test_reproject_return_footprint(ndcube_4d_ln_l_t_lt, wcs_4d_lt_t_l_ln):
    target_wcs_header = wcs_4d_lt_t_l_ln.low_level_wcs.to_header()
    target_wcs_header['CDELT3'] = 0.1   # original value = 0.2
    target_wcs = astropy.wcs.WCS(header=target_wcs_header)
    shape_out = (5, 20, 12, 8)

    resampled_cube, footprint = ndcube_4d_ln_l_t_lt.reproject_to(target_wcs, shape_out=shape_out,
                                                                 return_footprint=True)

    assert ndcube_4d_ln_l_t_lt.data.shape == (5, 10, 12, 8)
    assert resampled_cube.data.shape == (5, 20, 12, 8)
    assert footprint.shape == (5, 20, 12, 8)


def test_reproject_shape_out(ndcube_4d_ln_l_t_lt, wcs_4d_lt_t_l_ln):
    # should raise an exception when neither shape_out is specified nor
    # target_wcs has the pixel_shape or array_shape attribute
    wcs_4d_lt_t_l_ln.pixel_shape = None
    with pytest.raises(Exception):
        _ = ndcube_4d_ln_l_t_lt.reproject_to(wcs_4d_lt_t_l_ln)
    # should not raise an exception when shape_out is specified
    shape = (5, 10, 12, 8)
    _ = ndcube_4d_ln_l_t_lt.reproject_to(wcs_4d_lt_t_l_ln, shape_out=shape)

    # should not raise an exception when target_wcs has pixel_shape or array_shape attribute
    wcs_4d_lt_t_l_ln.array_shape = shape
    _ = ndcube_4d_ln_l_t_lt.reproject_to(wcs_4d_lt_t_l_ln, shape_out=shape)


def test_reproject_adaptive(ndcube_2d_ln_lt, wcs_2d_lt_ln):
    shape_out = (10, 12)
    resampled_cube = ndcube_2d_ln_lt.reproject_to(wcs_2d_lt_ln, algorithm='adaptive',
                                                  shape_out=shape_out)

    assert ndcube_2d_ln_lt.data.shape == (10, 12)
    assert resampled_cube.data.shape == (10, 12)


def test_reproject_exact(ndcube_2d_ln_lt, wcs_2d_lt_ln):
    shape_out = (10, 12)
    resampled_cube = ndcube_2d_ln_lt.reproject_to(wcs_2d_lt_ln, algorithm='exact',
                                                  shape_out=shape_out)

    assert ndcube_2d_ln_lt.data.shape == (10, 12)
    assert resampled_cube.data.shape == (10, 12)


def test_reproject_invalid_algorithm(ndcube_4d_ln_l_t_lt, wcs_4d_lt_t_l_ln):
    with pytest.raises(ValueError):
        _ = ndcube_4d_ln_l_t_lt.reproject_to(wcs_4d_lt_t_l_ln, algorithm='my_algorithm',
                                             shape_out=(5, 10, 12, 8))


def test_reproject_adaptive_incompatible_wcs(ndcube_4d_ln_l_t_lt, wcs_4d_lt_t_l_ln,
                                             wcs_1d_l, ndcube_1d_l):
    with pytest.raises(ValueError):
        _ = ndcube_1d_l.reproject_to(wcs_1d_l, algorithm='adaptive',
                                     shape_out=(10,))

    with pytest.raises(ValueError):
        _ = ndcube_4d_ln_l_t_lt.reproject_to(wcs_4d_lt_t_l_ln, algorithm='adaptive',
                                             shape_out=(5, 10, 12, 8))


def test_reproject_exact_incompatible_wcs(ndcube_4d_ln_l_t_lt, wcs_4d_lt_t_l_ln,
                                          wcs_1d_l, ndcube_1d_l):
    with pytest.raises(ValueError):
        _ = ndcube_1d_l.reproject_to(wcs_1d_l, algorithm='exact',
                                     shape_out=(10,))

    with pytest.raises(ValueError):
        _ = ndcube_4d_ln_l_t_lt.reproject_to(wcs_4d_lt_t_l_ln, algorithm='exact',
                                             shape_out=(5, 10, 12, 8))


############### REBIN ############################

@pytest.mark.parametrize("bin_shape", [(10, 2, 1), (-1, 2, 1)])
def test_rebin(ndcube_3d_l_ln_lt_ectime, bin_shape):
    cube = ndcube_3d_l_ln_lt_ectime[:, 1:]
    with pytest.warns(NDCubeUserWarning, match="The uncertainty on this NDCube has no known way to propagate forward"):
        output = cube.rebin(bin_shape, operation=np.sum, propagate_uncertainties=True)
    output_sc, output_spec = output.axis_world_coords(wcs=output.wcs)
    output_time, = output.axis_world_coords(wcs=output.extra_coords)

    # Build expected cube contents.
    expected_data = np.array([[3840, 3860, 3880, 3900, 3920, 3940, 3960, 3980],
                              [4160, 4180, 4200, 4220, 4240, 4260, 4280, 4300]])
    expected_mask = np.zeros(expected_data.shape, dtype=bool)
    expected_uncertainty = None
    expected_unit = cube.unit
    expected_meta = cube.meta
    expected_Tx = np.array([[9.99999999, 19.99999994, 29.99999979, 39.9999995,
                             49.99999902, 59.99999831, 69.99999731, 79.99999599],
                            [9.99999999, 19.99999994, 29.99999979, 39.9999995,
                             49.99999902, 59.99999831, 69.99999731, 79.99999599]]) * u.arcsec
    expected_Ty = np.array([[-12.49999997, -12.49999993, -12.49999985, -12.49999975,
                             -12.49999962, -12.49999946, -12.49999926, -12.49999904],
                            [-2.5       ,  -2.49999999,  -2.49999997,  -2.49999995,
                             -2.49999993,  -2.49999989,  -2.49999986,  -2.49999981]]) * u.arcsec
    expected_spec = SpectralCoord([1.11e-09], unit=u.m)
    expected_time = Time([51544.00104167, 51544.00243056], format="mjd", scale="utc")
    expected_time.format = "fits"

    # Confirm output is as expected.
    assert np.all(output.shape == np.array([1, 2, 8]))
    assert np.all(output.data == expected_data)
    assert np.all(output.mask == expected_mask)
    assert output.uncertainty == expected_uncertainty
    assert output.unit == expected_unit
    assert output.meta == expected_meta
    assert u.allclose(output_sc.Tx, expected_Tx)
    assert u.allclose(output_sc.Ty, expected_Ty)
    assert u.allclose(output_spec, expected_spec)
    assert output_time.scale == expected_time.scale
    assert output_time.format == expected_time.format
    assert np.allclose(output_time.mjd, expected_time.mjd)


def test_rebin_dask(ndcube_2d_dask):
    da = pytest.importorskip("dask.array")
    # Execute rebin.
    cube = ndcube_2d_dask
    bin_shape = (4, 2)
    output = cube.rebin(bin_shape, propagate_uncertainties=True)
    dask_type = da.core.Array
    assert isinstance(output.data, dask_type)
    assert isinstance(output.uncertainty.array, dask_type)
    assert isinstance(output.mask, dask_type)


def test_rebin_bin_shape_quantity(ndcube_3d_l_ln_lt_ectime):
    # Confirm rebin's bin_shape argument handles being a astropy unit
    cube = ndcube_3d_l_ln_lt_ectime[:, 1:]
    cube._extra_coords = ExtraCoords(cube)
    bin_shape = (10, 2, 1) * u.pix
    output = cube.rebin(bin_shape)
    np.testing.assert_allclose(output.shape, cube.shape / bin_shape.to_value())
    with pytest.raises(u.UnitConversionError, match=re.escape("'m' (length) and 'pix' are not convertible")):
        cube.rebin((10, 2, 1) * u.m)


def test_rebin_no_ec(ndcube_3d_l_ln_lt_ectime):
    # Confirm rebin does not try to handle extra coords when there aren't any.
    cube = ndcube_3d_l_ln_lt_ectime[:, 1:]
    cube._extra_coords = ExtraCoords(cube)
    bin_shape = (10, 2, 1)
    with pytest.warns(NDCubeUserWarning, match="The uncertainty on this NDCube has no known way to propagate forward"):
        output = cube.rebin(bin_shape, operation=np.mean, propagate_uncertainties=True)
    assert output.extra_coords.is_empty


def test_rebin_uncerts(ndcube_2d_ln_lt_uncert):
    cube = ndcube_2d_ln_lt_uncert
    bin_shape = (2, 4)
    output = cube.rebin(bin_shape, operation=np.mean, propagate_uncertainties=True)
    output_uncert = output.uncertainty.array
    expected_uncert = (np.array([[2.73495887,  3.68239053,  4.7116876],
                                 [9.07524104, 10.1882285, 11.30486621],
                                 [15.79240324, 16.91744662, 18.0432813],
                                 [22.55216176, 23.68037162, 24.80886938],
                                 [29.32507459, 30.45455631, 31.58417325]])
                       / np.array(bin_shape).prod())
    assert np.allclose(output_uncert, expected_uncert)


def test_rebin_some_masked_uncerts(ndcube_2d_ln_lt_mask_uncert):
    cube = ndcube_2d_ln_lt_mask_uncert
    bin_shape = (2, 4)
    expected_data = np.array([[7.5,  11.5,  15.5],
                              [31.5,  35.5,  39.5],
                              [55.5,  59.5,  63.5],
                              [79.5,  83.5,  87.5],
                              [103.5, 107.5, 111.5]])
    expected_uncert = np.array([[0.34186986, 0.46029882, 0.58896095],
                                [1.13440513, 1.27352856, 1.41310828],
                                [1.9740504, 2.11468083, 2.25541016],
                                [2.81902022, 2.96004645, 3.10110867],
                                [3.66563432, 3.80681954, 3.94802166]])
    expected_mask = np.zeros((5, 3), dtype=bool)
    expected_mask[:3, 0] = True
    # Execute function and assert result is as expected.
    output = cube.rebin(bin_shape, operation=np.mean, operation_ignores_mask=True,
                        handle_mask=np.any, propagate_uncertainties=True)
    assert np.allclose(output.data, expected_data)
    assert np.allclose(output.uncertainty.array[np.logical_not(expected_mask)],
                       expected_uncert[np.logical_not(expected_mask)])
    assert (output.mask == expected_mask).all()


def test_rebin_some_masked_uncerts_exclude_masked_values(ndcube_2d_ln_lt_mask_uncert):
    cube = ndcube_2d_ln_lt_mask_uncert
    bin_shape = (2, 4)
    expected_data = np.array([[6.71428571,  11.5,  15.5],
                              [31.5,  35.5,  39.5],
                              [0.,  59.5,  63.5],
                              [79.5,  83.5,  87.5],
                              [103.5, 107.5, 111.5]])
    expected_uncert = np.array([[0.34374884, 0.46029882, 0.58896095],
                                [1.30586285, 1.27352856, 1.41310828],
                                [0., 2.11468083, 2.25541016],
                                [2.81902022, 2.96004645, 3.10110867],
                                [3.66563432, 3.80681954, 3.94802166]])
    expected_mask = np.zeros((5, 3), dtype=bool)
    expected_mask[2, 0] = True
    # Execute function and assert result is as expected.
    output = cube.rebin(bin_shape, operation=np.mean, operation_ignores_mask=False,
                        handle_mask=np.all, propagate_uncertainties=True)
    assert np.allclose(output.data, expected_data)
    assert np.allclose(output.uncertainty.array[np.logical_not(expected_mask)],
                       expected_uncert[np.logical_not(expected_mask)])
    assert (output.mask == expected_mask).all()


def test_rebin_errors(ndcube_3d_l_ln_lt_ectime):
    cube = ndcube_3d_l_ln_lt_ectime
    # Wrong number of axes in bin_shape)
    with pytest.raises(ValueError, match="bin_shape must have an entry for each"):
        cube.rebin((2,))

    # bin shape shouldn't have any negatives
    with pytest.raises(ValueError, match="bin_shape should not be less than -1"):
        cube.rebin((2, -2, 1))
    # bin_shape not integer multiple of data shape.
    with pytest.raises(ValueError, match="bin_shape must be an integer fraction"):
        cube.rebin((9, 2, 1))

    # bin_shape larger than data shape.
    with pytest.raises(ValueError, match="bin_shape cannot be larger than data shape"):
        cube.rebin((20, 2, 1))


def test_rebin_no_propagate(ndcube_2d_ln_lt_mask_uncert):
    # Execute rebin.
    cube = ndcube_2d_ln_lt_mask_uncert
    bin_shape = (2, 4)

    cube._mask[:] = True
    with pytest.warns(NDCubeUserWarning, match="Uncertainties cannot be propagated as all values are masked and operation_ignores_mask is False."):
        output = cube.rebin(bin_shape, operation=np.sum, propagate_uncertainties=True,
                            operation_ignores_mask=False)
    assert output.uncertainty is None

    cube._mask = True
    with pytest.warns(NDCubeUserWarning, match="Uncertainties cannot be propagated as all values are masked and operation_ignores_mask is False."):
        output = cube.rebin(bin_shape, operation=np.sum, propagate_uncertainties=True,
                            operation_ignores_mask=False)
    assert output.uncertainty is None

    cube._mask = False
    cube._uncertainty = UnknownUncertainty(cube.data * 0.1)
    with pytest.warns(NDCubeUserWarning, match="The uncertainty on this NDCube has no known way to propagate forward"):
        output = cube.rebin(bin_shape, operation=np.sum, propagate_uncertainties=True)
    assert output.uncertainty is None


def test_rebin_axis_aware_meta(ndcube_4d_axis_aware_meta):
    # Execute rebin.
    cube = ndcube_4d_axis_aware_meta
    bin_shape = (1, 2, 5, 1)
    output = cube.rebin(bin_shape, operation=np.sum)

    # Build expected meta
    expected_meta = copy.deepcopy(cube.meta)
    del expected_meta._axes["pixel label"]
    del expected_meta._axes["line"]
    expected_meta._data_shape = np.array([5, 4, 2, 12], dtype=int)

    # Confirm output meta is as expected.
    helpers.assert_metas_equal(output.meta, expected_meta)


def test_rebin_specutils():
    pytest.importorskip("specutils")
    # Tests for https://github.com/sunpy/ndcube/issues/717
    try:
        from specutils import Spectrum  # noqa PLC0415
    except ImportError:
        from specutils import Spectrum1D as Spectrum  # noqa PLC0415
    y = np.arange(4000)*u.ct
    x = np.arange(200, 4200)*u.nm
    spec = Spectrum(flux=y, spectral_axis=x, bin_specification='centers', mask=x > 2000*u.nm)
    output = spec.rebin((10,), operation=np.sum, operation_ignores_mask=False)
    assert output.shape == (400,)



import inspect
from functools import wraps
from itertools import chain

import numpy as np

import astropy.nddata
from astropy.wcs.wcsapi import BaseHighLevelWCS, BaseLowLevelWCS, HighLevelWCSWrapper, SlicedLowLevelWCS

from ndcube.utils import wcs as wcs_utils
from ndcube.utils.exceptions import warn_user

__all__ = [
    "get_crop_item_from_points",
    "propagate_rebin_uncertainties",
    "sanitize_crop_inputs",
    "sanitize_wcs",
]


def sanitize_wcs(func):
    """
    A wrapper for NDCube methods to sanitise the wcs argument.

    This decorator is only designed to be used on methods of NDCube.

    It will find the wcs argument, keyword or positional and if it is `None`, set
    it to ``self.wcs``.
    It will then verify that the WCS has a matching number of pixel dimensions
    to the dimensionality of the array. It will finally verify that the object
    passed is a HighLevelWCS object, or an ExtraCoords object.
    """
    # This needs to be here to prevent a circular import
    from ndcube.extra_coords.extra_coords import ExtraCoords  # noqa: PLC0415

    @wraps(func)
    def wcs_wrapper(*args, **kwargs):
        sig = inspect.signature(func)
        params = sig.bind(*args, **kwargs)
        wcs = params.arguments.get('wcs', None)
        self = params.arguments['self']

        if wcs is None:
            wcs = self.wcs

        if not isinstance(wcs, ExtraCoords):
            if not wcs.pixel_n_dim == self.data.ndim:
                raise ValueError("The supplied WCS must have the same number of "
                                 "pixel dimensions as the NDCube object. "
                                 "If you specified `cube.extra_coords.wcs` "
                                 "please just pass `cube.extra_coords`.")

        if not isinstance(wcs, (BaseHighLevelWCS, ExtraCoords)):
            raise TypeError("wcs argument must be a High Level WCS or an ExtraCoords object.")

        params.arguments['wcs'] = wcs

        return func(*params.args, **params.kwargs)

    return wcs_wrapper


def sanitize_crop_inputs(points, wcs):
    """Sanitize inputs to NDCube crop methods.

    First arg returned signifies whether the inputs imply that cropping
    should be performed or not.
    """
    points = list(points)
    n_points = len(points)
    n_coords = [None] * n_points
    values_are_none = [False] * n_points
    for i, point in enumerate(points):
        # Ensure each point is a list
        if isinstance(point, (tuple, list)):
            points[i] = list(point)
        else:
            points[i] = [point]
        # Record number of objects in each point.
        # Later we will ensure all points have same number of objects.
        n_coords[i] = len(points[i])
        # Confirm whether point contains at least one None entry.
        if all(coord is None for coord in points[i]):
            values_are_none[i] = True
        # Squeeze length-1 coordinate objects to scalars.
        points[i] = [coord.squeeze() if hasattr(coord, "squeeze") else coord for coord in points[i]]
    # If no points contain a coord, i.e. if all entries in all points are None,
    # set no-op flag to True and exit.
    if all(values_are_none):
        return True, points, wcs
    # Not not all points are of same length, error.
    if len(set(n_coords)) != 1:
        raise ValueError("All points must have same number of coordinate objects."
                         f"Number of objects in each point: {n_coords}")
    # Import must be here to avoid circular import.
    from ndcube.extra_coords.extra_coords import ExtraCoords  # noqa: PLC0415
    if isinstance(wcs, ExtraCoords):
        # Determine how many dummy axes are needed
        n_dummy_axes = len(wcs._cube_array_axes_without_extra_coords)
        if n_dummy_axes > 0:
            points = [point + [None] * n_dummy_axes for point in points]
        # Convert extra coords to WCS describing whole cube.
        wcs = wcs.cube_wcs
    # Ensure WCS is low level.
    if isinstance(wcs, BaseHighLevelWCS):
        wcs = wcs.low_level_wcs
    return False, points, wcs


def get_crop_item_from_points(points, wcs, crop_by_values, keepdims, original_shape):
    """
    Find slice item that crops to minimum cube in array-space containing specified world points.

    Parameters
    ----------
    points : iterable of iterables
        Each iterable represents a point in real world space.
        Each element in a point gives the real world coordinate value of the point
        in high-level coordinate objects or quantities.
        (Must be consistently high or low level within and across points.)
        Objects must be in the order required by
        wcs.world_to_array_index/world_to_array_index_values.

    wcs : `~astropy.wcs.wcsapi.BaseHighLevelWCS`, `~astropy.wcs.wcsapi.BaseLowLevelWCS`
        The WCS to use to convert the world coordinates to array indices.

    crop_by_values : `bool`
        Denotes whether cropping is done using high-level objects or "values",
        i.e. low-level objects.

    keepdims : `bool`
        If `False`, returned item will drop length-1 dimensions otherwise, item will keep length-1 dimensions.

    original_shape: `tuple` of `int`
        The shape of the data cube before cropping.

    Returns
    -------
    item : `tuple` of `slice`
        The slice item for each axis of the cube which, when applied to the cube,
        will return the minimum cube in array-index-space that contains all the
        input world points.
    """
    # Define a list of lists to hold the pixel coordinates of the points
    # where each inner list gives the pixel coordinates of all points for that pixel axis.
    # Recall that pixel axis ordering is reversed compared to array axis ordering.
    combined_points_pixel_idx = [[]] * wcs.pixel_n_dim
    high_level_wcs = HighLevelWCSWrapper(wcs) if isinstance(wcs, BaseLowLevelWCS) else wcs
    low_level_wcs = high_level_wcs.low_level_wcs
    # For each point compute the corresponding array indices.
    for point in points:
        # Get the pixel axes associated with each element in point.
        point_inputs_pixel_axes = (
            tuple(wcs_utils.world_axis_to_pixel_axes(i, low_level_wcs.axis_correlation_matrix)
                  for i in range(low_level_wcs.world_n_dim)) if crop_by_values
            else wcs_utils.pixel_indices_for_world_objects(high_level_wcs))
        # Get indices of pixel axes which correspond to only None inputs in point
        # as well as those that correspond to a coord.
        point_indices_with_inputs = []
        pixel_axes_with_input = []
        for i, coord in enumerate(point):
            if coord is not None:
                point_indices_with_inputs.append(i)
                pixel_axes_with_input.append(point_inputs_pixel_axes[i])
        pixel_axes_with_input = set(chain.from_iterable(pixel_axes_with_input))
        pixel_axes_without_input = set(range(low_level_wcs.pixel_n_dim)) - pixel_axes_with_input
        pixel_axes_with_input = np.array(list(pixel_axes_with_input))
        pixel_axes_without_input = np.array(list(pixel_axes_without_input))
        # Slice out the axes that do not correspond to a coord
        # from the WCS and the input point.
        if len(pixel_axes_without_input) > 0:
            array_axes_without_input = wcs_utils.convert_between_array_and_pixel_axes(
                pixel_axes_without_input, low_level_wcs.pixel_n_dim)
            wcs_slice = np.array([slice(None)] * low_level_wcs.pixel_n_dim)
            wcs_slice[array_axes_without_input] = 0
            sliced_wcs = SlicedLowLevelWCS(low_level_wcs, slices=tuple(wcs_slice))
            sliced_point = np.array(point, dtype=object)[np.array(point_indices_with_inputs)]
        else:
            # Else, if all axes have at least one crop input, no need to slice the WCS.
            sliced_wcs, sliced_point = low_level_wcs, np.array(point, dtype=object)
        # Derive the pixel indices of the input point and place each index
        # in the list corresponding to its axis.
        # Use the to_pixel methods to preserve fractional indices for future rounding.
        if crop_by_values:
            # world_to_pixel_values is APE14 low-level API and expects plain
            # floats, not Quantity objects. So we need to strip units here; the values are
            # already in the correct units because _get_crop_by_values_item called
            # .to(wcs.world_axis_units[j]) before reaching this point.
            #
            # Passing Quantity objects raises TypeError in gWCS when the WCS's
            # declared high-level type is itself Quantity (e.g., a WCS built from
            # QuantityTableCoordinate), because gWCS cannot distinguish such inputs
            # from an accidental high-level API call.
            stripped_point = [p.value if hasattr(p, "value") else p for p in sliced_point]
            point_pixel_indices = sliced_wcs.world_to_pixel_values(*stripped_point)
        else:
            point_pixel_indices = HighLevelWCSWrapper(sliced_wcs).world_to_pixel(*sliced_point)
        # For each pixel axis associated with this point, place the pixel coords for
        # that pixel axis into the corresponding list within combined_points_pixel_idx.
        if sliced_wcs.pixel_n_dim == 1:
            point_pixel_indices = (point_pixel_indices,)
        for axis, index in zip(pixel_axes_with_input, point_pixel_indices):
            combined_points_pixel_idx[axis] = combined_points_pixel_idx[axis] + [index]

    # Iterate through each array axis to determine the min and max pixel coords
    # and then convert to array indices. Note that combined_points_pixel_idx holds the
    # pixel coords for each pixel axis. Therefore, to iterate in array axis order,
    # combined_points_pixel_idx must be reversed.
    item = []
    ambiguous = False
    message = ""
    result_is_scalar = True
    for array_axis, pixel_coords in enumerate(combined_points_pixel_idx[::-1]):
        if pixel_coords == []:
            result_is_scalar = False
            item.append(slice(None))
        else:
            # Calculate the index of the array element containing the pixel coordinate.
            # Note that integer pixel coordinates correspond to the pixel center,
            # while integer array indices correspond to lower edge of desired array element.
            # Therefore a shift of 0.5 is required in the conversion.
            # The max idx conversion below will discard right-ward array element if
            # max pixel coord corresponds to a pixel edge.
            min_array_idx = int(np.floor(min(pixel_coords) + 0.5))
            max_array_idx = int(np.ceil(max(pixel_coords) - 0.5)) + 1
            # Raise error if indices all lie below or all lie above array axis's extent.
            # Exception: min_array_idx == max_array_idx == 0 is allowed because max_array_idx
            # will be later changed to 1.
            if (min_array_idx < 0 and max_array_idx <= 0) or min_array_idx >= original_shape[array_axis]:
                raise ValueError(f"All world points associated with array axis {array_axis}"
                                 " are outside the range of the NDCube being cropped.")
            # world_to_array_index uses negative indices to represent locations to the left
            # of the 0th pixel, while python slicing uses them to count backwards from the
            # last element in the array. Therefore, set negative indices to 0.
            # Note that we've already checked that the max pixel_coord is >= 0.
            # Also note that there's no need to clip the max array idx, as values above
            # the array extent does not cause ambiguity in the slicing so long as the
            # min array idx is below that upper extent, which has also already been checked
            # by the above error.
            if min_array_idx < 0:
                min_array_idx = 0
            # Due to the above calculation, the above min and max array indices can only be
            # same if the original pixel coords correspond to the same pixel edge.
            # If this is the case, increment the max array index by 1 so the rightward array
            # element is kept. Also, build a warning message about this to be raised later.
            if min_array_idx == max_array_idx:
                ambiguous = True
                max_array_idx += 1
                if min_array_idx == 0:
                    message += (f"All input points corresponding to array axis {array_axis} lie on "
                                "the lower boundary of array element 0 (the first element). "
                                "The cropped NDCube will only include array element 0.\n")
                else:
                    message += (f"All input points corresponding to array axis {array_axis} lie on "
                                f"the boundary between array elements {min_array_idx - 1} and "
                                f"{min_array_idx}. The cropped NDCube will only include array "
                                f"element {min_array_idx}.\n")
            if max_array_idx - min_array_idx == 1 and not keepdims:
                item.append(min_array_idx)
            else:
                item.append(slice(min_array_idx, max_array_idx))
                result_is_scalar = False
    # Raise warning if all world values for any array axes correspond to a pixel edge.
    if ambiguous:
        warn_user(message)
    # If item will result in a scalar cube, raise an error as this is not currently supported.
    if result_is_scalar:
        raise ValueError("Input points causes cube to be cropped to a single pixel. "
                         "This is not supported when keepdims=False.")
    return tuple(item)


def propagate_rebin_uncertainties(uncertainty, data, mask, operation, operation_ignores_mask=False,
                                  propagation_operation=None, correlation=0, **kwargs):
    """
    Default algorithm for uncertainty propagation in :meth:`~ndcube.NDCube.rebin`.

    First dimension of uncertainty, data and mask inputs represent the pixels
    in the bin being aggregated by the rebin process while the latter dimensions
    must have the same shape as the rebinned data. The operation input is the
    function used to aggregate elements in the first dimension, e.g. `numpy.sum`.

    Parameters
    ----------
    uncertainty: `astropy.nddata.NDUncertainty`
        Cannot be instance of `astropy.nddata.UnknownUncertainty`.
        The uncertainties associated with the data. The first dimension represents
        pixels in each bin being aggregated while trailing dimensions must have
        the same shape as the rebinned data.
    data: array-like or `None`
        The data associated with the above uncertainties.
        Must have same shape as above.
    mask: array-like of `bool` or `None`
        Indicates whether any uncertainty elements should be ignored in propagation.
        If True, corresponding uncertainty element is ignored. If False, it is used.
        Must have same shape as above.
    operation: function
        The function used to aggregate the data for which the uncertainties are being
        propagated here.
    operation_ignores_mask: `bool`
        Determines whether masked values are used or excluded from calculation.
        Default is False causing masked data and uncertainty to be excluded.
    propagation_operation: function
        The operation which defines how the uncertainties are propagated.
        This can differ from operation, e.g. if operation is sum, then
        propagation_operation should be add.
    correlation: `int`
        Passed to `astropy.nddata.NDUncertainty.propagate`. See that method's docstring.
        Default=0.

    Returns
    -------
    new_uncertainty: `astropy.nddata.NDUncertainty`
        The propagated uncertainty.
        Same shape as input uncertainty without its first dimension.
    """
    flat_axis = 0
    operation_is_mean = True if operation in {np.mean, np.nanmean} else False
    operation_is_nantype = True if operation in {np.nansum, np.nanmean, np.nanprod} else False
    # If propagation_operation kwarg not set manually, try to set it based on operation kwarg.
    if not propagation_operation:
        if operation in {np.sum, np.nansum, np.mean, np.nanmean}:
            propagation_operation = np.add
        # TODO: product was renamed to prod for numpy 2.0
        elif operation in {np.prod, np.nanprod}:
            propagation_operation = np.multiply
        else:
            raise ValueError("propagation_operation not recognized.")
    # Build mask if not provided.
    new_uncertainty = uncertainty[0]  # Define uncertainty for initial iteration step.
    if operation_ignores_mask or mask is None:
        mask = False
    if mask is False:
        if operation_is_nantype:
            nan_mask = np.isnan(data)
            if nan_mask.any():
                mask = nan_mask
                idx = np.logical_not(mask)
                mask1 = mask[1:]
        else:
            # If there is no mask and operation is not nan-type, build generator
            # so non-mask can still be iterated.
            n_pix_per_bin = data.shape[flat_axis]
            mask1 = (False for i in range(1, n_pix_per_bin))
    else:
        # Mask uncertainties corresponding to nan data if operation is nantype.
        if operation_is_nantype:
            mask[np.isnan(data)] = True
        # Set masked uncertainties in first mask to 0
        # as they shouldn't count towards final uncertainty.
        mask1 = mask[1:]
        idx = np.logical_not(mask)
        uncertainty.array[mask] = 0
        new_uncertainty.array[mask[0]] = 0
    # Propagate uncertainties.
    # Note uncertainty must be associated with a parent nddata for some propagations.
    cumul_data = data[0]
    if mask is not False and operation_ignores_mask is False:
        cumul_data[idx[0]] = 0
    parent_nddata = astropy.nddata.NDData(cumul_data, uncertainty=new_uncertainty)
    new_uncertainty.parent_nddata = parent_nddata
    for j, mask_slice in enumerate(mask1):
        i = j + 1
        cumul_data = operation(data[:i+1]) if mask is False else operation(data[:i+1][idx[:i+1]])
        data_slice = astropy.nddata.NDData(data=data[i], mask=mask_slice,
                                           uncertainty=uncertainty[i])
        new_uncertainty = new_uncertainty.propagate(propagation_operation, data_slice,
                                                    cumul_data, correlation)
        parent_nddata = astropy.nddata.NDData(cumul_data, uncertainty=new_uncertainty)
        new_uncertainty.parent_nddata = parent_nddata
    # If aggregation operation is mean, uncertainties must be divided by
    # number of unmasked pixels in each bin.
    if operation_is_mean and propagation_operation is np.add:
        if mask is False:
            new_uncertainty.array /= n_pix_per_bin
        else:
            unmasked_per_bin = np.logical_not(mask).astype(int).sum(axis=flat_axis)
            new_uncertainty.array /= np.clip(unmasked_per_bin, 1, None)
    return new_uncertainty



import numbers

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import WCS
from astropy.wcs.wcsapi import HighLevelWCSWrapper

from ndcube.wcs.wrappers import ResampledLowLevelWCS


@pytest.fixture
def celestial_wcs(request):
    return request.getfixturevalue(request.param)


EXPECTED_2D_REPR_NUMPY2 = """
ResampledLowLevelWCS Transformation

This transformation has 2 pixel and 2 world dimensions

Array shape (Numpy order): (2, 15)

Pixel Dim  Axis Name  Data size  Bounds
        0  None              15  (-1.75, 13.25)
        1  None         2.33333  (0.0, 2.0)

World Dim  Axis Name        Physical Type  Units
        0  Right Ascension  pos.eq.ra      deg
        1  Declination      pos.eq.dec     deg

Correlation between pixel and world axes:

           Pixel Dim
World Dim    0    1
        0  yes  yes
        1  yes  yes
""".strip()
EXPECTED_2D_REPR_NUMPY1 = """
ResampledLowLevelWCS Transformation

This transformation has 2 pixel and 2 world dimensions

Array shape (Numpy order): (2.3333333333333335, 15.0)

Pixel Dim  Axis Name  Data size  Bounds
        0  None              15  (-1.75, 13.25)
        1  None         2.33333  (0.0, 2.0)

World Dim  Axis Name        Physical Type  Units
        0  Right Ascension  pos.eq.ra      deg
        1  Declination      pos.eq.dec     deg

Correlation between pixel and world axes:

           Pixel Dim
World Dim    0    1
        0  yes  yes
        1  yes  yes
""".strip()

@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs', 'celestial_2d_fitswcs'],
                         indirect=True)
def test_2d(celestial_wcs):

    # Upsample along the first pixel dimension and downsample along the second
    # pixel dimension.
    factors = [0.4, 3]
    wcs = ResampledLowLevelWCS(celestial_wcs, factors)

    # The following shouldn't change compared to the original WCS
    assert wcs.pixel_n_dim == 2
    assert wcs.world_n_dim == 2
    assert tuple(wcs.world_axis_physical_types) == ('pos.eq.ra', 'pos.eq.dec')
    assert tuple(wcs.world_axis_units) == ('deg', 'deg')
    assert tuple(wcs.pixel_axis_names) == ('', '')
    assert tuple(wcs.world_axis_names) == ('Right Ascension', 'Declination')
    assert_equal(wcs.axis_correlation_matrix, np.ones((2, 2)))

    # Shapes and bounds should be floating-point if needed
    assert_allclose(wcs.pixel_shape, (15, 7/3))
    assert_allclose(wcs.array_shape, (7/3, 15))
    assert_allclose(wcs.pixel_bounds, ((-1.75, 13.25), (0, 2)))

    under_pixel, over_pixel = (2.3, 1.2), (5.75+0.6*1.25, 0.2/1.5*0.5)
    world = celestial_wcs.pixel_to_world_values(*under_pixel)

    assert_allclose(wcs.pixel_to_world_values(*over_pixel), world)
    assert_allclose(wcs.array_index_to_world_values(*over_pixel[::-1]), world)
    assert_allclose(wcs.world_to_pixel_values(*world), over_pixel)
    assert_allclose(wcs.world_to_array_index_values(*world),
                    np.around(over_pixel[::-1]).astype(int))

    EXPECTED_2D_REPR = EXPECTED_2D_REPR_NUMPY2 if np.__version__ >= '2.0.0' else EXPECTED_2D_REPR_NUMPY1
    assert str(wcs) == EXPECTED_2D_REPR
    assert EXPECTED_2D_REPR in repr(wcs)

    celestial_wcs.pixel_bounds = None
    under_pixel_array = (np.array([2.3, 2.4]),
                         np.array([4.3, 4.4]))
    over_pixel_array = (np.array([over_pixel[0], 5.75+0.8*1.25]),
                        1 + np.array([0.3, 0.4]) / 1.5 * 0.5)
    world_array = celestial_wcs.pixel_to_world_values(*under_pixel_array)

    assert_allclose(wcs.pixel_to_world_values(*over_pixel_array), world_array)
    assert_allclose(wcs.array_index_to_world_values(*over_pixel_array[::-1]), world_array)
    assert_allclose(wcs.world_to_pixel_values(*world_array), over_pixel_array)
    assert_allclose(wcs.world_to_array_index_values(*world_array),
                    np.around(np.asarray(over_pixel_array)[::-1]).astype(int))

    wcs_hl = HighLevelWCSWrapper(wcs)

    celestial = wcs_hl.pixel_to_world(*over_pixel)
    assert isinstance(celestial, SkyCoord)
    assert_quantity_allclose(celestial.ra, world[0] * u.deg)
    assert_quantity_allclose(celestial.dec, world[1] * u.deg)

    celestial = wcs_hl.pixel_to_world(*over_pixel_array)
    assert isinstance(celestial, SkyCoord)
    assert_quantity_allclose(celestial.ra, world_array[0] * u.deg)
    assert_quantity_allclose(celestial.dec, world_array[1] * u.deg)


@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs',
                          'celestial_2d_fitswcs'],
                         indirect=True)
@pytest.mark.parametrize(('factor', 'offset', 'over_pixel'),
                         [(2, 0, (0.9, 1.9)),
                          (2, 1, (0.4, 1.4)),
                         ])
def test_scalar_factor_and_offset(celestial_wcs, factor, offset, over_pixel):
    celestial_wcs.pixel_bounds = None
    wcs = ResampledLowLevelWCS(celestial_wcs, factor, offset=offset)
    # Define the pixel coord pre-resampled pixel grid corresponding to
    # the same location in the resampled grid, as defined in the parameterization.
    under_pixel = (2.3, 4.3)
    # Get the corresponding world location using original WCS.
    world = celestial_wcs.pixel_to_world_values(*under_pixel)
    # Confirm resampled WCS maps to and from the same world coordinate to the
    # pixel location in the resample pixel coords.
    assert_allclose(wcs.pixel_to_world_values(*over_pixel), world)
    assert_allclose(wcs.array_index_to_world_values(*over_pixel[::-1]), world)
    assert_allclose(wcs.world_to_pixel_values(*world), over_pixel)
    assert_allclose(wcs.world_to_array_index_values(*world),
                    np.around(over_pixel[::-1]).astype(int))


@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs'],
                         indirect=True)
def test_factor_wrong_length_error(celestial_wcs):
    with pytest.raises(ValueError):
        ResampledLowLevelWCS(celestial_wcs, [2] * 3)


@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs'],
                         indirect=True)
def test_scalar_wrong_length_error(celestial_wcs):
    with pytest.raises(ValueError):
        ResampledLowLevelWCS(celestial_wcs, 2, offset=[1] * 3)


@pytest.mark.parametrize('celestial_wcs',
                         ['celestial_2d_ape14_wcs', 'celestial_2d_fitswcs'],
                         indirect=True)
def test_int_fraction_pixel_shape(celestial_wcs):
    # Some fractional factors are not representable by exact floats, e.g. 1/3.
    # However, it is still desirable for the pixel shape to return ints in these cases.
    # This test checks that this is the case.
    wcs = ResampledLowLevelWCS(celestial_wcs, 1/3)
    assert wcs.pixel_shape == (18, 21)
    for dim in wcs.pixel_shape:
        assert isinstance(dim, numbers.Integral)

@pytest.fixture
def four_five_wcs():
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = ["HPLN-TAN", "HPLT-TAN"]
    wcs.wcs.cdelt = [2, 2]
    wcs.wcs.crval = [0, 0]
    wcs.wcs.crpix = [1, 1]

    wcs.wcs.set()
    wcs.pixel_shape = [4, 5]
    return wcs

@pytest.mark.parametrize(('factor', 'offset', 'expected_over_pixels'),
                         [([2, 3], [0, 0], np.meshgrid(np.linspace(-0.5, 1.5, 4*2+1), np.linspace(-0.5, 1+1/6, 5*2+1))),
                          ([2, 3], [1, 2], np.meshgrid(np.linspace(-1, 1, 4*2+1), np.linspace(-1-1/6, 0.5, 5*2+1))),
                         ])
def test_resampled_pixel_to_world_values(four_five_wcs, factor, offset, expected_over_pixels):
    """
    Notes
    -----
    Below shows schematics of the two test cases tested in this test of how ResampledLowLevelWCS.
    The asterisks show the corners of pixels in a grid before resampling, while the dashes and
    pipes show the edges the resampled pixels. In the first case, the resampled pixels are
    obtained by applied resampling factors of 2 along the x-axis and 3 along the y-axis. No
    offset is applied to either axis. In the second case, the same resampling factors have
    been applied, but an offsets of 1 and 2 pixels have been applied to the x- and y-axes,
    respectively. The right column/upper row of numbers along the side of/below the grids denote
    the edges and centres of the original pixel grid in the original pixel coordinates.
    The left column and lower row gives the same locations in the pixel coordinates of the
    resampled grid.

    Test Case 1

    ::

        resampled  original
        factor=3
        offset=0

        1+1/6       4.5*         *         *         *         *
                       |                   |                   |
          1          4 |                   |                   |
                       |                   |                   |
         5/6        3.5*         *         *         *         *
                       |                   |                   |
         4/6         3 |                   |                   |
                       |                   |                   |
                       |                   |                   |
         0.5        2.5*---------*---------*---------*---------*
                       |                   |                   |
         2/6         2 |                   |                   |
                       |                   |                   |
         1/6        1.5*         *         *         *         *
                       |                   |                   |
          0          1 |                   |                   |
                       |                   |                   |
        -1/6        0.5*         *         *         *         *
                       |                   |                   |
        -2/6         0 |                   |                   |
                       |                   |                   |
        -0.5       -0.5*---------*---------*---------*---------*
                     -0.5   0   0.5   1   1.5   2   2.5   3   3.5  original pixel indices
                     -0.5 -0.25  0  0.25  0.5  0.75  1   1.25 1.5 resampled pixel indices: factor=2, offset=0

    Test Case 2

    ::

        resampled  original
        factor=3
        offset=2

          0.5      4.5 *-----------*-----------*-----------*-----------*
                                   |                       |
          2/6       4              |                       |
                                   |                       |
          1/6      3.5 *           *           *           *           *
                                   |                       |
           0        3              |                       |
                                   |                       |
         -1/3      2.5 *           *           *           *           *
                                   |                       |
         -2/6       2              |                       |
                                   |                       |
         -0.5      1.5 *-----------*-----------*-----------*-----------*
                                   |                       |
         -4/6       1              |                       |
                                   |                       |
         -5/6      0.5 *           *           *           *           *
                                   |                       |
          -1        0              |                       |
                                   |                       |
         -1-1/6   -0.5 *           *           *           *           *
                     -0.5    0    0.5    1    1.5    2    2.5    3    3.5  original pixel indices
                      -1   -0.75 -0.5  -0.25   0    0.25  0.5   0.75   1   resampled pixel indices: factor=2, offset=1
    """
    wcs = four_five_wcs
    # Get world values of original pixel grid.
    under_pixels = np.meshgrid(np.arange(-0.5, 4, 0.5), np.arange(-0.5, 5, 0.5))
    expected_world = wcs.pixel_to_world_values(*under_pixels)

    # Resample WCS
    new_wcs = ResampledLowLevelWCS(wcs, factor, offset)
    # Get expected pixel coords in resampled WCS of same pixel locations as above.
    output_world = new_wcs.pixel_to_world_values(*expected_over_pixels)
    assert_allclose(np.asarray(output_world), np.asarray(expected_world), atol=1e-15)



"""
=====================================
Changing the resolution of an NDCube
=====================================

This example shows how to change the sampling of an NDCube by reprojecting to a different grid.

Reprojecting is useful when you need a new WCS or a non-integer change in pixel scale.

.. warning::

    Reprojecting resamples the data, but it does not make the underlying data intrinsically higher resolution.
    If you only need to combine aligned pixels by an integer factor to reduce resolution,
    `~ndcube.NDCube.rebin` is usually a better choice because it operates directly on contiguous input pixels.
"""

import matplotlib.pyplot as plt

from astropy.io import fits
from astropy.wcs import WCS

from sunpy.data.sample import AIA_171_IMAGE
from sunpy.visualization.colormaps import cm

from ndcube import NDCube

############################################################################
# We start by creating an NDCube from sample solar data provided by SunPy.
# Here we use an AIA 171 image, but the same approach can be applied to other datasets, including those with non-celestial axes.

image_data = fits.getdata(AIA_171_IMAGE, ext=1)
image_header = fits.getheader(AIA_171_IMAGE, ext=1)
cube = NDCube(image_data, WCS(image_header))

###########################################################################
# Next, we define a new WCS with a finer pixel scale. This changes the sampling of the output array, but it does not make the underlying data intrinsically higher resolution.
# Note that while it is obvious that the CDELT values are changed to reflect the finer scale, the CRPIX values also need to be adjusted as the reference pixel position is different on the new pixel scale.
# You can scale the axes by any amount, including non-integer values, greater or less than 1.
# You can also scale each axis independently.
# Here we scale both spatial axes by a factor of 1.5.

scale_factor = 1.5

new_wcs = cube.wcs.deepcopy()
new_wcs.wcs.cdelt /= scale_factor
new_wcs.wcs.crpix *= scale_factor
new_shape = tuple(int(s * scale_factor) for s in cube.data.shape)

###########################################################################
# We reproject the original cube to the new WCS.
reprojected_cube = cube.reproject_to(new_wcs, shape_out=new_shape)

###########################################################################
# Our new NDCube now has finer sampling.
print(cube.data.shape, reprojected_cube.data.shape)

###########################################################################
# We can plot the original and reprojected cubes side by side.

fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(14, 8))

ax1.imshow(cube.data, origin='lower', cmap=cm.sdoaia171, vmin=10, vmax=10000)
ax1.set_title('Original')
ax1.set_xlabel('X [pixel]')
ax1.set_ylabel('Y [pixel]')

ax2.imshow(reprojected_cube.data, origin='lower', cmap=cm.sdoaia171, vmin=10, vmax=10000)
ax2.set_title('Reprojected to New Resolution')
ax2.set_xlabel('X [pixel]')
ax2.set_ylabel('Y [pixel]')
plt.tight_layout()
plt.show()



import pytest

import asdf
import astropy.units as u
from astropy.coordinates import SkyCoord

from ndcube.extra_coords import MultipleTableCoordinate


@pytest.fixture
def lut(request):
    return request.getfixturevalue(request.param)


def assert_table_coord_equal(test_table, expected_table):
    test_table = test_table.table
    expected_table = expected_table.table
    if not isinstance(expected_table, tuple):
        test_table = (test_table,)
        expected_table = (expected_table,)
    for test_tab, ex_tab in zip(test_table, expected_table):
        if ex_tab.isscalar:
            assert test_tab == ex_tab
        elif isinstance(ex_tab, SkyCoord):
            assert u.allclose(ex_tab.spherical.lat, test_tab.spherical.lat)
            assert u.allclose(ex_tab.spherical.lon, test_tab.spherical.lon)
        else:
            assert all(test_tab == ex_tab)


@pytest.mark.parametrize("lut",
                         [
                             "lut_1d_distance",
                             "lut_3d_distance_mesh",
                             "lut_1d_skycoord_no_mesh",
                             "lut_2d_skycoord_no_mesh",
                             "lut_2d_skycoord_mesh",
                             "lut_3d_skycoord_mesh",
                             "lut_1d_time",
                             "lut_1d_wave",
                         ], indirect=True)
def test_serialize(lut, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["lut"] = lut
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_table_coord_equal(af["lut"], lut)


def assert_mtc_equal(test_mtc, expected_mtc):
    assert len(test_mtc._table_coords) == len(expected_mtc._table_coords)
    assert len(test_mtc._dropped_coords) == len(expected_mtc._dropped_coords)

    for (test_tc, expected_tc) in zip(test_mtc._table_coords, expected_mtc._table_coords):
        assert_table_coord_equal(test_tc, expected_tc)

    for (test_tc, expected_tc) in zip(test_mtc._dropped_coords, expected_mtc._dropped_coords):
        assert_table_coord_equal(test_tc, expected_tc)


def test_serialize_multiple_coord(lut_1d_distance, lut_1d_time, tmp_path):
    mtc = MultipleTableCoordinate(lut_1d_distance, lut_1d_time)
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["lut"] = mtc
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        new_mtc = af["lut"]
        assert_mtc_equal(new_mtc, mtc)


def test_serialize_sliced_multiple_coord(lut_1d_distance, lut_1d_time, tmp_path):
    mtc = MultipleTableCoordinate(lut_1d_distance, lut_1d_time)[0, :]
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["lut"] = mtc
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        new_mtc = af["lut"]
        assert_mtc_equal(new_mtc, mtc)
